import ast
from padl.dumptools import packagefinder


source_a = '''
from bla import x, y
import blu as z
import ble.lo

def x():
    return 1
'''


def test_getpackages():
    out = packagefinder.get_packages(ast.parse(source_a).body)
    assert out == {'bla', 'blu', 'ble'}



import pytest
from padl import transform, Batchify
import padl.transforms as padl
from padl.util_transforms import IfTrain, IfEval, IfInfer


@transform
def plus_one(x):
    return x + 1


@transform
def times_two(x):
    return x * 2


times_three = transform(lambda x: x * 3)


class TestIfInStage:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = (
            plus_one
            >> IfTrain(times_two, plus_one)
            >> Batchify()
            >> times_three
        )
        request.cls.transform_2 = (
            plus_one
            >> IfEval(times_two)
            >> Batchify()
            >> times_three
        )
        request.cls.transform_3 = (
            plus_one
            >> IfInfer(times_two)
            >> Batchify()
            >> times_three
        )

    def test_infer_apply_1(self):
        assert self.transform_1.infer_apply(1) == 9

    def test_infer_apply_2(self):
        assert self.transform_2.infer_apply(1) == 6

    def test_infer_apply_3(self):
        assert self.transform_3.infer_apply(1) == 12

    def test_eval_apply_1(self):
        assert list(self.transform_1.eval_apply([1, 2])) == [9, 12]

    def test_eval_apply_2(self):
        assert list(self.transform_2.eval_apply([1, 2])) == [12, 18]

    def test_eval_apply_3(self):
        assert list(self.transform_3.eval_apply([1, 2])) == [6, 9]

    def test_train_apply_1(self):
        assert list(self.transform_1.train_apply([1, 2])) == [12, 18]

    def test_train_apply_2(self):
        assert list(self.transform_2.train_apply([1, 2])) == [6, 9]

    def test_train_apply_3(self):
        assert list(self.transform_3.train_apply([1, 2])) == [6, 9]

    def test_save_and_load_1(self, tmp_path):
        self.transform_1.pd_save(tmp_path / 'test.padl')
        t1 = padl.load(tmp_path / 'test.padl')
        assert t1.infer_apply(1) == 9

    def test_save_and_load_2(self, tmp_path):
        self.transform_2.pd_save(tmp_path / 'test.padl')
        t2 = padl.load(tmp_path / 'test.padl')
        assert t2.infer_apply(1) == 6

    def test_save_and_load_3(self, tmp_path):
        self.transform_3.pd_save(tmp_path / 'test.padl')
        t3 = padl.load(tmp_path / 'test.padl')
        assert t3.infer_apply(1) == 12



from collections import OrderedDict
import pytest
import torch
from padl import transforms as pd, transform, Identity
from padl.transforms import Batchify, Unbatchify
from padl.dumptools.serialize import value
from collections import namedtuple
from padl.exceptions import WrongDeviceError


@transform
def plus_one(x):
    return x + 1


@transform
def append_one(x):
    return x + "one"


@transform
def times_two(x):
    return x * 2


@transform
def plus(x, y):
    return x + y


@transform
def get_info(x):
    return x['info']


@transform
def complex_signature_func_1(a, b=10):
    return a+b


@transform
def complex_signature_func_2(*a, b= 10):
    return sum(a) + b


def simple_func(x):
    return x


class SimpleClass:
    def __init__(self, a):
        self.a = a

    def __call__(self, x):
        return x + self.a


@transform
class SimpleClassTransform:
    def __init__(self, a):
        self.a = a

    def __call__(self, x):
        return x + self.a


@transform
class ClassTransformWithManyArguments:
    def __init__(self, a, b, *args, c=1, d=2, **kwargs):
        self.a = a

    def __call__(self, x):
        return x + self.a


@transform
def trans_with_globals(x, y):
    return (plus >> times_two)(x, y)


@transform
class ClassLookup:
    def __init__(self, dic):
        self.dic = dic

    def __call__(self, args):
        return [self.dic.get(x, len(self.dic)) for x in args]


@transform
class Polynomial(torch.nn.Module):
    def __init__(self, a, b):
        super().__init__()
        self.a = torch.nn.Parameter(torch.tensor(float(a)))
        self.b = torch.nn.Parameter(torch.tensor(float(b)))

    def forward(self, x):
        return x**self.a + x**self.b


class PolynomialClass(torch.nn.Module):
    def __init__(self, a, b):
        super().__init__()
        self.a = torch.nn.Parameter(torch.tensor(float(a)))
        self.b = torch.nn.Parameter(torch.tensor(float(b)))

    def forward(self, x):
        return x**self.a + x**self.b


def test_isinstance_of_namedtuple():
    tup = tuple([1, 2, 3])

    namedtup_type = namedtuple('something', 'a b c')
    namedtup_ins = namedtup_type(*tup)

    assert pd._isinstance_of_namedtuple(namedtup_ins)
    assert not pd._isinstance_of_namedtuple(tup)
    assert not pd._isinstance_of_namedtuple(list(tup))
    assert not pd._isinstance_of_namedtuple(1.)
    assert not pd._isinstance_of_namedtuple('something')


class TestLFCallTransform:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = plus_one >> (times_two + times_two)
        request.cls.transform_2 = transform(simple_func) + transform(simple_func) \
            + transform(simple_func)
        request.cls.transform_3 = plus_one + times_two >> plus
        request.cls.transform_4 = plus_one + times_two >> complex_signature_func_1
        request.cls.transform_5 = plus_one >> complex_signature_func_1
        request.cls.transform_6 = plus_one + times_two >> complex_signature_func_2

    def test_infer_apply(self):
        assert self.transform_1.infer_apply(1)
        assert self.transform_2.infer_apply(10)
        assert self.transform_3.infer_apply(1.4)
        assert self.transform_4.infer_apply(201)
        assert self.transform_5.infer_apply(11.1)
        assert self.transform_6.infer_apply(19)

    def test_save_load(self, tmp_path):
        for transform_ in [self.transform_1,
                           self.transform_2,
                           self.transform_3,
                           self.transform_4,
                           self.transform_5,
                           self.transform_6]:
            transform_.pd_save(tmp_path / 'test.padl', True)
            t_ = pd.load(tmp_path / 'test.padl')
            assert t_.infer_apply(1)


class TestMap:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = ~plus_one
        request.cls.transform_2 = transform(simple_func) / ~plus_one
        request.cls.transform_3 = ~times_two + ~plus_one
        request.cls.transform_4 = transform(lambda x: [x, x, x]) >> ~plus_one

    def test_pd_preprocess(self):
        assert isinstance(self.transform_1.pd_preprocess, pd.Identity)
        assert isinstance(self.transform_2.pd_preprocess, pd.Identity)

    def test_pd_forward(self):
        assert isinstance(self.transform_1.pd_forward, pd.Map)
        assert isinstance(self.transform_2.pd_forward, pd.Parallel)

    def test_pd_postprocess(self):
        assert isinstance(self.transform_1.pd_postprocess, pd.Identity)
        assert isinstance(self.transform_2.pd_postprocess, pd.Identity)

    def test_infer_apply(self):
        assert self.transform_1.infer_apply([2, 3, 4]) == [3, 4, 5]
        assert self.transform_2.infer_apply((1, [2, 3, 4])) == (1, [3, 4, 5])
        assert self.transform_3.infer_apply([2, 3, 4]) == ([4, 6, 8], [3, 4, 5])
        assert self.transform_4.infer_apply(1) == [2, 2, 2]

    def test_eval_apply(self):
        assert list(self.transform_1.eval_apply([[2, 3], [3, 4]])) == [[3, 4], [4, 5]]
        assert list(self.transform_2.eval_apply(([1, [2, 3]], (2, [3, 4])))) == [(1, [3, 4]),
                                                                                 (2, [4, 5])]
        assert list(self.transform_3.eval_apply([[2, 3], [2, 3]])) == \
               [([4, 6], [3, 4]), ([4, 6], [3, 4])]
        assert list(self.transform_4.eval_apply([1])) == [[2, 2, 2]]

    def test_train_apply(self):
        assert list(self.transform_1.train_apply([[2, 3], [3, 4]])) == [[3, 4], [4, 5]]
        assert list(self.transform_2.train_apply(([1, [2, 3]], (2, [3, 4])))) == [(1, [3, 4]),
                                                                                  (2, [4, 5])]
        assert list(self.transform_3.train_apply([[2, 3], [2, 3]])) == \
               [([4, 6], [3, 4]), ([4, 6], [3, 4])]
        assert list(self.transform_4.train_apply([1])) == [[2, 2, 2]]

    def test_save_and_load(self, tmp_path):
        self.transform_1.pd_save(tmp_path / 'test.padl')
        t1 = pd.load(tmp_path / 'test.padl')
        assert t1.infer_apply([2, 3, 4]) == [3, 4, 5]
        self.transform_2.pd_save(tmp_path / 'test.padl', True)
        t2 = pd.load(tmp_path / 'test.padl')
        assert t2.infer_apply((1, [2, 3, 4])) == (1, [3, 4, 5])
        self.transform_3.pd_save(tmp_path / 'test.padl', True)
        t3 = pd.load(tmp_path / 'test.padl')
        assert t3.infer_apply([2, 3, 4]) == ([4, 6, 8], [3, 4, 5])
        self.transform_4.pd_save(tmp_path / 'test.padl', True)
        t4 = pd.load(tmp_path / 'test.padl')
        assert t4.infer_apply(1) == [2, 2, 2]


class TestParallel:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = plus_one / times_two / times_two
        request.cls.transform_2 = transform(simple_func) / transform(simple_func) / transform(simple_func)
        request.cls.transform_3 = plus_one / plus_one / transform(simple_func)

    def test_output(self):
        in_ = (2, 2, 2)
        out = self.transform_1(in_)
        assert pd._isinstance_of_namedtuple(out)
        assert out._fields == ('plus_one', 'times_two_0', 'times_two_1')

        out = self.transform_2(in_)
        assert pd._isinstance_of_namedtuple(out)
        assert out._fields == ('out_0', 'out_1', 'out_2')

        out = self.transform_3(in_)
        assert pd._isinstance_of_namedtuple(out)
        assert out._fields == ('plus_one_0', 'plus_one_1', 'out_2')

    def test_pd_preprocess(self):
        assert isinstance(self.transform_1.pd_preprocess, pd.Identity)

    def test_pd_forward(self):
        assert isinstance(self.transform_1.pd_forward, pd.Parallel)

    def test_pd_postprocess(self):
        assert isinstance(self.transform_1.pd_postprocess, pd.Identity)

    def test_infer_apply(self):
        assert self.transform_1.infer_apply((2, 3, 4)) == (3, 6, 8)

    def test_eval_apply(self):
        assert list(self.transform_1.eval_apply([(2, 3, 4), (3, 3, 4)])) == [(3, 6, 8), (4, 6, 8)]

    def test_context(self):
        assert self.transform_1.pd_stage is None
        with self.transform_1.pd_set_stage('train'):
            assert self.transform_1.pd_stage is 'train'
            assert self.transform_1.pd_preprocess.pd_stage == 'train'
            assert self.transform_1.pd_forward.pd_stage == 'train'
            assert self.transform_1.pd_postprocess.pd_stage == 'train'

    def test_save_and_load(self, tmp_path):
        self.transform_1.pd_save(tmp_path / 'test.padl')
        t1 = pd.load(tmp_path / 'test.padl')
        assert t1.infer_apply((2, 3, 4)) == (3, 6, 8)
        self.transform_2.pd_save(tmp_path / 'test.padl', True)
        _ = pd.load(tmp_path / 'test.padl')
        self.transform_3.pd_save(tmp_path / 'test.padl', True)
        _ = pd.load(tmp_path / 'test.padl')


class TestRollout:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = plus_one + times_two + times_two
        request.cls.transform_2 = transform(simple_func) + transform(simple_func) + transform(simple_func)
        request.cls.transform_3 = plus_one + plus_one + transform(simple_func)

    def test_output(self):
        in_ = 123
        out = self.transform_1(in_)
        assert pd._isinstance_of_namedtuple(out)
        assert out._fields == ('plus_one', 'times_two_0', 'times_two_1')

        out = self.transform_2(in_)
        assert pd._isinstance_of_namedtuple(out)
        assert out._fields == ('out_0', 'out_1', 'out_2')

        out = self.transform_3(in_)
        assert pd._isinstance_of_namedtuple(out)
        assert out._fields == ('plus_one_0', 'plus_one_1', 'out_2')

    def test_pd_preprocess(self):
        assert isinstance(self.transform_1.pd_preprocess, pd.Identity)

    def test_pd_forward(self):
        assert isinstance(self.transform_1.pd_forward, pd.Rollout)

    def test_pd_postprocess(self):
        assert isinstance(self.transform_1.pd_postprocess, pd.Identity)

    def test_infer_apply(self):
        assert self.transform_1.infer_apply(2) == (3, 4, 4)

    def test_eval_apply(self):
        assert list(self.transform_1.eval_apply([2, 3])) == [(3, 4, 4), (4, 6, 6)]

    def test_context(self):
        assert self.transform_1.pd_stage is None
        with self.transform_1.pd_set_stage('train'):
            assert self.transform_1.pd_stage is 'train'
            assert self.transform_1.pd_preprocess.pd_stage == 'train'
            assert self.transform_1.pd_forward.pd_stage == 'train'
            assert self.transform_1.pd_postprocess.pd_stage == 'train'

    def test_save_and_load(self, tmp_path):
        self.transform_1.pd_save(tmp_path / 'test.padl')
        t1 = pd.load(tmp_path / 'test.padl')
        assert t1.infer_apply(2) == (3, 4, 4)
        self.transform_2.pd_save(tmp_path / 'test.padl', True)
        _ = pd.load(tmp_path / 'test.padl')
        self.transform_3.pd_save(tmp_path / 'test.padl', True)
        _ = pd.load(tmp_path / 'test.padl')


class TestCompose:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = (plus_one >> times_two) >> times_two
        request.cls.transform_2 = plus_one >> (times_two >> times_two)
        request.cls.transform_3 = plus_one >> times_two >> times_two
        request.cls.transform_4 = plus_one >> plus_one >> plus_one
        request.cls.transform_5 = (
            plus_one
            >> Batchify()
            >> times_two - 'named_times_two'
            >> times_two
            >> Unbatchify()
        )

    def test_associative(self):
        in_ = 123
        assert self.transform_1(in_) == self.transform_2(in_) == self.transform_3(in_)

    def test_output(self):
        assert self.transform_4(1) == 4

    def test_pd_preprocess(self):
        assert isinstance(self.transform_1.pd_preprocess, pd.Identity)
        assert isinstance(self.transform_5.pd_preprocess, pd.Compose)

    def test_pd_forward(self):
        assert isinstance(self.transform_1.pd_forward, pd.Compose)
        assert isinstance(self.transform_5.pd_forward, pd.Compose)

    def test_pd_postprocess(self):
        assert isinstance(self.transform_1.pd_postprocess, pd.Identity)
        assert isinstance(self.transform_5.pd_postprocess, pd.Unbatchify)

    def test_infer_apply(self):
        assert self.transform_4.infer_apply(1) == 4
        assert self.transform_5.infer_apply(1) == torch.tensor(8)

    def test_eval_apply(self):
        assert list(self.transform_5.eval_apply([1, 1])) == [torch.tensor([8]), torch.tensor([8])]

    def test_train_apply(self):
        # default
        assert list(self.transform_5.train_apply([1, 1])) == [torch.tensor([8]), torch.tensor([8])]
        # loader kwargs
        for batch in list(self.transform_5.train_apply(
            [1, 2, 1, 2],
            verbose=True,
            batch_size=2)
        ):
            assert torch.all(batch == torch.tensor([8, 12]))
        # flatten = True
        assert list(self.transform_5.train_apply(
            [1, 2, 1, 2],
            flatten=True,
            verbose=True,
            batch_size=2)
        ) == [torch.tensor([8]), torch.tensor([12]), torch.tensor([8]), torch.tensor([12])]

    def test_context(self):
        assert self.transform_1.pd_stage is None
        with self.transform_1.pd_set_stage('eval'):
            assert self.transform_1.pd_stage is 'eval'
            assert self.transform_1.pd_preprocess.pd_stage == 'eval'
            assert self.transform_1.pd_forward.pd_stage == 'eval'
            assert self.transform_1.pd_postprocess.pd_stage == 'eval'

    def test_all_transforms_1(self):
        c = plus_one >> times_two >> times_two
        all_ = c._pd_all_transforms()
        assert set(all_) == set([plus_one, times_two, c])

    def test_all_transforms_2(self):
        c = plus_one >> times_two >> trans_with_globals
        all_ = c._pd_all_transforms()
        assert set(all_) == set([plus_one, times_two, c, trans_with_globals, plus])

    def test_save_and_load(self, tmp_path):
        self.transform_1.pd_save(tmp_path / 'test.padl')
        _ = pd.load(tmp_path / 'test.padl')
        self.transform_2.pd_save(tmp_path / 'test.padl', True)
        _ = pd.load(tmp_path / 'test.padl')
        self.transform_3.pd_save(tmp_path / 'test.padl', True)
        _ = pd.load(tmp_path / 'test.padl')
        self.transform_4.pd_save(tmp_path / 'test.padl', True)
        t4 = pd.load(tmp_path / 'test.padl')
        assert t4.infer_apply(1) == 4
        self.transform_5.pd_save(tmp_path / 'test.padl', True)
        t5 = pd.load(tmp_path / 'test.padl')
        assert t5.infer_apply(1) == torch.tensor(8)

    def test_getitem(self):
        assert isinstance(self.transform_5[0], pd.Transform)
        assert isinstance(self.transform_5[0:2], pd.CompoundTransform)
        assert isinstance(self.transform_5[0:2], pd.Compose)
        assert isinstance(self.transform_5['named_times_two'], pd.Transform)
        with pytest.raises(ValueError):
            self.transform_5['other_name']
        with pytest.raises(TypeError):
            self.transform_5[2.1]


class TestModel:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        transform_1 = (
            plus_one
            >> Batchify()
            >> times_two
            >> Unbatchify()
            >> plus_one
        )
        transform_2 = (
            plus_one
            >> Batchify()
            >> times_two
            >> Unbatchify()
            >> plus_one
        ) - 'transform_2'
        request.cls.model_1 = (transform_1 / transform_2)
        request.cls.model_2 = (transform_1 + transform_2)
        request.cls.model_3 = (
            plus_one + times_two
            >> plus_one / Batchify()
            >> Batchify() / times_two
        )
        request.cls.model_4 = (
            plus_one + times_two
            >> Batchify()
            >> Polynomial(1, 0) / Identity()
            >> Unbatchify()
            >> plus_one / times_two
        ) - 'model_4'

    def test_pd_preprocess(self):
        assert isinstance(self.model_1.pd_preprocess, pd.Parallel)
        assert isinstance(self.model_2.pd_preprocess, pd.Rollout)
        assert isinstance(self.model_4.pd_preprocess, pd.Compose)

    def test_pd_forward(self):
        assert isinstance(self.model_1.pd_forward, pd.Parallel)
        assert isinstance(self.model_2.pd_forward, pd.Parallel)

    def test_pd_postprocess(self):
        assert isinstance(self.model_1.pd_postprocess, pd.Parallel)
        assert isinstance(self.model_2.pd_postprocess, pd.Parallel)
        assert isinstance(self.model_4.pd_postprocess, pd.Compose)

    def test_infer_apply(self):
        assert self.model_1.infer_apply((5, 5)) == (13, 13)
        assert self.model_2.infer_apply(torch.tensor(5)) == (13, 13)
        assert self.model_3.infer_apply(5) == (7, 20)
        assert self.model_4.infer_apply(5) == (8, 20)

    def test_eval_apply(self):
        assert list(self.model_1.eval_apply([(5, 5), (5, 5)])) == [(13, 13), (13, 13)]
        assert list(self.model_2.eval_apply(torch.tensor([5, 5]))) == [(13, 13), (13, 13)]
        assert list(self.model_3.eval_apply([5, 6])) == [(7, 20), (8, 24)]
        assert list(self.model_4.eval_apply([5, 6])) == [(8, 20), (9, 24)]

    def test_train_apply(self):
        assert list(self.model_1.train_apply([(5, 5), (5, 5)])) == [(13, 13), (13, 13)]
        assert list(self.model_2.train_apply(torch.tensor([5, 5]))) == [(13, 13), (13, 13)]
        assert list(self.model_3.train_apply([5, 6])) == [(7, 20), (8, 24)]
        assert list(self.model_4.train_apply([5, 6])) == [(8, 20), (9, 24)]

    def test_save_and_load(self, tmp_path):
        pd.save(self.model_1, tmp_path / 'test.padl')
        m1 = pd.load(tmp_path / 'test.padl')
        assert m1.infer_apply((5, 5)) == (13, 13)
        self.model_2.pd_save(tmp_path / 'test.padl', True)
        m2 = pd.load(tmp_path / 'test.padl')
        assert m2.infer_apply(5) == (13, 13)
        self.model_3.pd_save(tmp_path / 'test.padl', True)
        m3 = pd.load(tmp_path / 'test.padl')
        assert m3.infer_apply(5) == (7, 20)
        self.model_4.pd_save(tmp_path / 'test.padl', True)
        m4 = pd.load(tmp_path / 'test.padl')
        assert m4.infer_apply(5) == (8, 20)


class TestFunctionTransform:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = plus_one
        request.cls.transform_2 = get_info

    def test_pd_preprocess(self):
        assert isinstance(self.transform_1.pd_preprocess, pd.Identity)

    def test_pd_forward(self):
        assert isinstance(self.transform_1.pd_forward, pd.FunctionTransform)

    def test_pd_postprocess(self):
        assert isinstance(self.transform_1.pd_postprocess, pd.Identity)

    def test_infer_apply(self):
        assert self.transform_1.infer_apply(5) == 6

    def test_eval_apply(self):
        out = list(self.transform_1.eval_apply([5, 6]))
        assert len(out) == 2
        assert out[0] == 6
        assert out[1] == 7

        out = list(self.transform_2.eval_apply([{'info': 'hello'}, {'info': 'dog'}]))
        assert len(out) == 2
        assert out[0] == 'hello'
        assert out[1] == 'dog'

    def test_context(self):
        assert self.transform_1.pd_stage is None
        with self.transform_1.pd_set_stage('infer'):
            assert self.transform_1.pd_stage is 'infer'
            assert self.transform_1.pd_preprocess.pd_stage == 'infer'
            assert self.transform_1.pd_forward.pd_stage == 'infer'
            assert self.transform_1.pd_postprocess.pd_stage == 'infer'

    def test_all_transforms(self):
        all_ = trans_with_globals._pd_all_transforms()
        assert set(all_) == set([plus, times_two, trans_with_globals])

    def test_pd_to(self):
        self.transform_1.pd_to('cpu')
        assert self.transform_1.pd_device == 'cpu'

    def test_save_and_load(self, tmp_path):
        self.transform_1.pd_save(tmp_path / 'test.padl', True)
        t1 = pd.load(tmp_path / 'test.padl')
        assert t1.infer_apply(5) == 6
        self.transform_2.pd_save(tmp_path / 'test.padl', True)
        _ = pd.load(tmp_path / 'test.padl')


def test_name():
    assert (plus_one - 'p1')._pd_name == 'p1'
    assert plus_one._pd_name is None


class TestTransformDeviceCheck:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = (plus_one >> times_two) >> times_two
        request.cls.transform_2 = plus_one >> (times_two >> times_two)

    def test_device_check(self):
        self.transform_1.pd_to('gpu')
        self.transform_1.transforms[1].pd_to('cpu')

        with pytest.raises(WrongDeviceError):
            self.transform_1._pd_forward_device_check()

        self.transform_2.pd_to('gpu')
        assert self.transform_2._pd_forward_device_check()


class TestClassTransform:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = transform(SimpleClass)(2)
        request.cls.transform_2 = SimpleClassTransform(2)
        dic = {s: i for i, s in enumerate('abcdefghijklmnop')}
        request.cls.transform_3 = ClassLookup(dic=value(dic))
        request.cls.dic = dic

    def test_infer_apply(self):
        assert self.transform_1.infer_apply(1) == 3
        assert self.transform_2.infer_apply(1) == 3
        assert self.transform_3.infer_apply('abc') == [0, 1, 2]

    def test_save_and_load(self, tmp_path):
        self.transform_1.pd_save(tmp_path / 'test.padl')
        t1 = pd.load(tmp_path / 'test.padl')
        assert t1.infer_apply(1) == 3
        self.transform_2.pd_save(tmp_path / 'test.padl', True)
        t2 = pd.load(tmp_path / 'test.padl')
        assert t2.infer_apply(1) == 3
        self.transform_3.pd_save(tmp_path / 'test.padl', True)
        t3 = pd.load(tmp_path / 'test.padl')
        assert t3.dic == self.dic
        assert t3.infer_apply('abc') == [0, 1, 2]

    def test_stored_arguments(self):
        c = ClassTransformWithManyArguments(1, 2, 3, 4, 5)
        assert c._pd_arguments == OrderedDict([('a', 1), ('b', 2), ('args', (3, 4, 5))])


class TestTorchModuleTransform:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = Polynomial(2, 3)

    def test_output(self):
        output = self.transform_1(1)
        assert output == 2

    def test_infer_apply(self):
        assert self.transform_1.infer_apply(1)

    def test_device(self):
        self.transform_1.pd_to('cpu')
        device = next(self.transform_1.pd_layers[0].parameters()).device.type
        assert device == 'cpu'

    def test_pd_layers(self):
        assert len(self.transform_1.pd_layers) > 0

    def test_pd_parameters(self):
        params = list(self.transform_1.pd_parameters())
        assert len(params) == 2

    def test_save_and_load(self, tmp_path):
        self.transform_1.pd_save(tmp_path / 'test.padl')
        t1 = pd.load(tmp_path / 'test.padl')
        assert t1.infer_apply(1) == 2


class TestClassInstance:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = transform(SimpleClass(1))
        request.cls.transform_2 = transform(PolynomialClass(1, 2))

    def test_wrap(self):
        assert isinstance(self.transform_1, SimpleClass)
        assert isinstance(self.transform_1, pd.Transform)
        assert isinstance(self.transform_2, PolynomialClass)
        assert isinstance(self.transform_2, pd.Transform)

    def test_infer_apply(self):
        assert self.transform_1.infer_apply(1) == 2
        assert self.transform_2.infer_apply(1) == 2

    def test_eval_apply(self):
        assert list(self.transform_1.eval_apply([1])) == [2]
        assert list(self.transform_2.eval_apply([2])) == [6]

    def test_train_apply(self):
        assert list(self.transform_1.train_apply([1])) == [2]
        assert list(self.transform_2.train_apply([2])) == [6]

    def test_print(self):
        assert str(self.transform_1)
        assert str(self.transform_2)

    def test_pd_layers(self):
        assert len(self.transform_2.pd_layers) > 0

    def test_pd_parameters(self):
        params = list(self.transform_2.pd_parameters())
        assert len(params) == 2

    def test_save_and_load(self, tmp_path):
        self.transform_1.pd_save(tmp_path / 'test.padl')
        t1 = pd.load(tmp_path / 'test.padl')
        assert t1.infer_apply(1) == 2
        self.transform_2.pd_save(tmp_path / 'test.padl', True)
        t2 = pd.load(tmp_path / 'test.padl')
        assert t2.infer_apply(1) == 2



from padl.dumptools import sourceget


a = \
'''xxxx
01234567890
01234567890xxxxxxxx
01234567890xxx
01234567890
xxx
'''

a_r = \
'''xxxx
01234567890
01here567890xxxxxxxx
01234567890xxx
01234567890
xxx
'''

a_r1 = \
'''xxxx
01234567890
01he
re890xxxxxxxx
01234567890xxx
01234567890
xxx
'''

b = '0123456789'

b_r = '0123here6789'


class TestReplace:
    def test_a(self):
        assert sourceget._replace(a, 'here', 2, 2, 2, 5) == a_r

    def test_a1(self):
        assert sourceget._replace(a, 'he\nre', 2, 2, 2, 8) == a_r1

    def test_b(self):
        assert sourceget._replace(b, 'here', 0, 0, 4, 6) == b_r



import ast
import builtins
from collections import Counter, namedtuple
from collections.abc import Iterable
from dataclasses import dataclass
import sys
from typing import Optional

from padl.dumptools import symfinder

try:
    unparse = ast.unparse
except AttributeError:  # python < 3.9
    from astunparse import unparse


class Finder(ast.NodeVisitor):
    """Class for finding ast nodes of a given type in an ast tree.

    Example:

    >>> f = Finder(ast.Name).find(ast.parse('x(y)'))
    [<_ast.Name at ...>, <_ast.Name at ...>]
    """

    def __init__(self, nodetype):
        self.nodetype = nodetype
        self.result = []

    def generic_visit(self, node):
        if isinstance(node, self.nodetype):
            self.result.append(node)
        for child_node in ast.iter_child_nodes(node):
            child_node.parent = node
        super().generic_visit(node)

    def find(self, node):
        """Find subnodes of node. """
        self.visit(node)
        return self.result

    def get_source_segments(self, source):
        nodes = self.find(ast.parse(source))
        return [
            (ast.get_source_segment(source, node),
             (node.lineno, node.end_lineno, node.col_offset, node.end_col_offset)
            )
            for node in nodes
        ]


Vars = namedtuple('Vars', 'globals locals')


class _VarFinder(ast.NodeVisitor):
    """NodeVisitor that traverses all nodes subnodes and finds all named things. """

    def __init__(self):
        super().__init__()
        self.globals = set()
        self.locals = set()

    def find(self, node):
        """Find all globals and locals.

        :param node: An ast node to search.
        :returns: Tuple of sets with names of globals and locals.
        """
        if isinstance(node, list):
            for node_ in node:
                self.find(node_)
            return Vars(self.globals, self.locals)
        if isinstance(node, ast.FunctionDef):
            return self._find_in_function_def(node)
        self.visit(node)
        return Vars(self.globals, self.locals)

    def _find_in_function_def(self, node):
        """Special case: exclude args from globals. """
        for arg in node.args.args + node.args.posonlyargs + node.args.kwonlyargs:
            self.locals.add(arg.arg)
        if node.args.vararg is not None:
            self.locals.add(node.args.vararg.arg)
        if node.args.kwarg is not None:
            self.locals.add(node.args.kwarg.arg)
        for n in ast.iter_child_nodes(node):
            self.visit(n)
        return Vars(self.globals, self.locals)

    def visit_Name(self, node):
        if node.id not in self.locals:
            self.globals.add(node.id)

    def visit_withitem(self, node):
        self.visit(node.context_expr)
        if node.optional_vars is not None:
            self.locals.add(node.optional_vars.id)

    def visit_Assign(self, node):
        for target in node.targets:
            self.locals.update([x.id for x in Finder(ast.Name).find(target)])
        self.visit(node.value)

    def visit_For(self, node):
        self.locals.update([x.id for x in Finder(ast.Name).find(node.target)])
        for child in node.body:
            self.visit(child)

    def visit_NamedExpr(self, node):
        self.locals.update(Finder(ast.Name).find(node.target))

    def visit_comprehension(self, node):
        """Special case for comprehension - comprehension targets should be ignored. """
        targets = set()
        for gen in node.generators:
            for name in Finder(ast.Name).find(gen.target):
                targets.add(name.id)
        all_ = set(x.id for x in Finder(ast.Name).find(node))
        self.globals.update(all_ - targets - self.locals)

    def visit_DictComp(self, node):
        self.visit_comprehension(node)

    def visit_ListComp(self, node):
        self.visit_comprehension(node)

    def visit_SetComp(self, node):
        self.visit_comprehension(node)

    def visit_GeneratorExp(self, node):
        pass

    def visit_Lambda(self, node):
        for arg in node.args.args + node.args.posonlyargs + node.args.kwonlyargs:
            self.locals.add(arg.arg)
        self.visit(node.body)

    def visit_FunctionDef(self, node):
        self.locals.add(node.name)
        inner_globals = find_globals(node)
        self.globals.update(inner_globals - self.locals)

    def visit_ClassDef(self, node):
        self.locals.add(node.name)
        inner_globals = find_globals(node)
        self.globals.update(inner_globals - self.locals)

    def visit_Import(self, node):
        for name in node.names:
            if name.asname is None:
                self.locals.add(name.name)
            else:
                self.locals.add(name.asname)

    def visit_ImportFrom(self, node):
        for name in node.names:
            if name.asname is None:
                self.locals.add(name.name)
            else:
                self.locals.add(name.asname)


class _Renamer(ast.NodeTransformer):
    def __init__(self, from_, to, rename_locals):
        self.from_ = from_
        self.to = to
        self.rename_locals = rename_locals

    def visit_arg(self, node):
        if node.arg == self.from_:
            return ast.arg(**{**node.__dict__, 'arg': self.to})
        return node

    def visit_Name(self, node):
        if node.id == self.from_:
            return ast.Name(**{**node.__dict__, 'id': self.to})
        return node

    def visit_FunctionDef(self, node):
        if self.rename_locals and node.name == self.from_:
            name = self.to
        else:
            name = node.name
        if not self.rename_locals:
            globals_ = find_globals(node)
            if self.from_ not in globals_:
                return node
        return ast.FunctionDef(**{**node.__dict__,
                                  'body': rename(node.body, self.from_, self.to),
                                  'name': name})

    def visit_ClassDef(self, node):
        if self.rename_locals and node.name == self.from_:
            name = self.to
        else:
            name = node.name
        if not self.rename_locals:
            globals_ = find_globals(node)
            if self.from_ not in globals_:
                return node
        return ast.ClassDef(**{**node.__dict__,
                               'body': rename(node.body, self.from_, self.to),
                               'name': name})


def rename(tree, from_, to, rename_locals=False):
    if not rename_locals:
        globals_ = find_globals(tree)
        if from_ not in globals_:
            return tree
    if isinstance(tree, Iterable) and not isinstance(tree, str):
        return [rename(x, from_, to) for x in tree]
    elif not isinstance(tree, ast.AST):
        return tree
    renamer = _Renamer(from_, to, rename_locals)
    return renamer.visit(tree)


class _MethodFinder(ast.NodeVisitor):
    """Find all methods in a class node. """

    def __init__(self):
        self.methods = {}

    def visit_FunctionDef(self, node):
        self.methods[node.name] = node

    def find(self, node: ast.ClassDef):
        self.visit(node)
        return self.methods


def _find_globals_in_classdef(node: ast.ClassDef, filter_builtins=True):
    """Find globals used below a ClassDef node. """
    methods = _MethodFinder().find(node)
    # globals used in class body
    globals_ = set()
    for child in ast.iter_child_nodes(node):
        globals_.update(_VarFinder().find(child).globals)
    # globals used in methods
    for method in methods.values():
        globals_.update(_VarFinder().find(method).globals)
    if filter_builtins:
        globals_ = {x for x in globals_ if x not in builtins.__dict__}
    return globals_


def find_globals(node: ast.AST, filter_builtins=True):
    """Find all globals used below a node. """
    if isinstance(node, ast.ClassDef):
        return _find_globals_in_classdef(node)
    globals_ = _VarFinder().find(node).globals
    if filter_builtins:
        globals_ = {x for x in globals_ if x not in builtins.__dict__}
    return globals_


def build_codegraph(x: set, module=None, all_=None, update_callback=None):
    if module is None:
        module = sys.modules['__main__']
    if all_ is None:
        all_ = {}
    todo = set(x)
    while todo and (next_ := todo.pop()):
        if next_ in all_:
            continue
        if update_callback is not None and update_callback(next_, module):
            pass
        else:
            source, node = symfinder.find(next_, module)
            globals_ = find_globals(node)
            all_[next_] = (source, globals_, node)
        todo.update(globals_)
    return all_


def _get_nodes_without_in_edges(graph):
    """Get all nodes in directed graph *graph* that don't have incoming edges.

    The graph is represented by a dict mapping nodes to incoming edges.
    """
    nextlevel = set()
    for node, deps in graph.items():
        if not deps:
            nextlevel.add(node)
    filtered_graph = {}
    for node, deps in graph.items():
        if node in nextlevel:
            continue
        filtered_graph[node] = \
            {dep for dep in deps if dep not in nextlevel}
    return nextlevel, filtered_graph


def _topsort(graph):
    """Topologically sort a graph represented by a dict mapping nodes to incoming edges.
    Careful: Doesn't check for loops.
    """
    levels = []
    while graph:
        nextlevel, graph = _get_nodes_without_in_edges(graph)
        levels.append(nextlevel)
    return levels


# sort precedence
_PRECEDENCE = {
    ast.Import: lambda k, v: '1' + v[2].names[0].name.lower() + k.lower(),
    ast.ImportFrom: lambda k, v: '1' + v[2].module.lower() + k.lower(),
    ast.Assign: lambda k, _v: '2' + k.lower()
}


def _sort(graph):
    top = _topsort({k: v[1] for k, v in graph.items()})

    def sortkey(x):
        val = graph[x]
        try:
            return _PRECEDENCE[val[2].__class__](x, val)
        except KeyError:
            return 'zz' + x.lower()

    res = []
    for level in top:
        res += sorted(level, key=sortkey)
    return res


def unscope_graph(graph, scopemap):
    """Create a version of *graph* where all non-top level variables are renamed (by prepending
    the scope) to prevent conflicts."""
    counts = Counter(x[0] for x in graph)
    to_rename = set(k for k, c in counts.items() if c > 1)
    scopemap = {**scopemap}
    scopemap.update({(k[0], v): v for k, v in scopemap})
    def unscope(name, scope):
        if name in to_rename:
            return scope.unscoped(name)
        return name
    res = {}
    for k, v in graph.items():
        changed = False
        k_unscoped = unscope(*k)
        changed = changed or k_unscoped != k[0]
        code = v.source
        tree = ast.parse(code)
        rename(tree, k[0], k_unscoped, rename_locals=True)
        vars_ = set()
        for var in list(v.globals_):
            var_unscoped = unscope(var[0], scopemap[var])
            changed = changed or var_unscoped != var[0]
            rename(tree, var[0], var_unscoped)
            vars_.add(var_unscoped)
        if changed:
            code = unparse(tree).strip('\n')
        res[k_unscoped] = (code, vars_, v.ast_node)
    return res


@dataclass
class CodeNode:
    source: str
    globals_: set
    ast_node: Optional[ast.AST] = None

    @classmethod
    def from_source(cls, source, scope):
        node = ast.parse(source).body[0]
        globals_ = {
            (var, scope)
            for var in find_globals(node)
        }

        return cls(
            source=source,
            ast_node=node,
            globals_=globals_
        )


def dumps_graph(graph):
    sorted_ = _sort(graph)
    res = ''
    for i, name in enumerate(sorted_):
        here = graph[name]
        res += here[0]
        if i < len(sorted_) - 1:
            next_ = graph[sorted_[i + 1]]
            if isinstance(here[2], (ast.Import, ast.ImportFrom)) \
                    and isinstance(next_[2], (ast.Import, ast.ImportFrom)):
                res += '\n'
            elif isinstance(here[2], ast.Assign) \
                    and isinstance(next_[2], ast.Assign):
                res += '\n'
            else:
                res += '\n\n\n'
    return res + '\n'


def dumps(x: str, module=None):
    graph = build_codegraph({x}, module)
    return dumps_graph(graph)



import ast

from padl.dumptools import symfinder


def test__get_call_signature_a():
    source = 'a(x, y, z)'
    signature = symfinder._get_call_signature(source)
    assert signature == (['x', 'y', 'z'], {})


def test__get_call_signature_b():
    source = 'a(x, y, z=1)'
    signature = symfinder._get_call_signature(source)
    assert signature == (['x', 'y'], {'z': '1'})


def test__get_call_assignments_a():
    source = 'def f(x, y, z=1):pass'
    args = ast.parse(source).body[0].args
    values, keywords = symfinder._get_call_signature('f(1, 2, 3)')
    assignments = symfinder._get_call_assignments(args, source, values, keywords)
    assert assignments == {'x': '1', 'y': '2', 'z': '3'}


def test__get_call_assignments_b():
    source = 'def f(x, y, z=1):pass'
    args = ast.parse(source).body[0].args
    values, keywords = symfinder._get_call_signature('f(1, 2)')
    assignments = symfinder._get_call_assignments(args, source, values, keywords)
    assert assignments == {'x': '1', 'y': '2', 'z': '1'}


def test__get_call_assignments_c():
    source = 'def f(*args, z=1):pass'
    args = ast.parse(source).body[0].args
    values, keywords = symfinder._get_call_signature('f(1, 2, 3, z=4)')
    assignments = symfinder._get_call_assignments(args, source, values, keywords)
    assert assignments == {'args': str([1, 2, 3]), 'z': '4'}


def test__get_call_assignments_d():
    source = 'def f(a, b, *args, z=1):pass'
    args = ast.parse(source).body[0].args
    values, keywords = symfinder._get_call_signature('f(1, 2, 3, z=4)')
    assignments = symfinder._get_call_assignments(args, source, values, keywords)
    assert assignments == {'a': '1', 'b': '2', 'args': '[3]', 'z': '4'}


def test__get_call_assignments_e():
    source = 'def f(a, b, *args, z=1, **kwargs):pass'
    args = ast.parse(source).body[0].args
    values, keywords = symfinder._get_call_signature('f(1, 2, 3, z=4, u=7, f=8)')
    assignments = symfinder._get_call_assignments(args, source, values, keywords)
    assert assignments == {'a': '1',
                           'b': '2',
                           'args': '[3]',
                           'z': '4',
                           'kwargs': "{'u': 7, 'f': 8}"}


def test__get_call_assignments_f():
    source = 'def f(a, b, /, c):pass'
    args = ast.parse(source).body[0].args
    values, keywords = symfinder._get_call_signature('f(1, 2, 3)')
    assignments = symfinder._get_call_assignments(args, source, values, keywords)
    assert assignments == {'a': '1',
                           'b': '2',
                           'c': '3'}



"""Extra useful Transforms. """

from collections import OrderedDict
from typing import Optional

from padl.transforms import ClassTransform, Identity, Transform, Stage


class IfInStage(ClassTransform):
    """Perform *if_* if called in stage *target_stage*, else perform *else_*.

    Example:

        >>> a = transform(lambda x: x + 10)
        >>> b = transform(lambda x: x * 10)
        >>> iis = IfInStage(a, 'infer', b)
        >>> iis.infer_apply(1)
        11
        >>> list(iis.eval_apply([1]))
        [100]

    :param if_: Transform to apply when the stage matches.
    :param target_stage: Stage (one of 'train', 'eval', 'infer').
    :param else_: Transform to apply when the stage doesn't match (defaults to identity transform).
    """

    pd_dont_dump_code = True

    def __init__(self, if_: Transform, target_stage: Stage, else_: Optional[Transform] = None):
        super().__init__(arguments=OrderedDict([('if_', if_),
                                                ('target_stage', target_stage),
                                                ('else_', else_)]))

        assert target_stage in ('train', 'eval', 'infer'), "Target stage can only be train, " \
                                                           "eval or infer"

        if else_ is None:
            else_ = Identity()

        self.if_ = if_
        self.else_ = else_
        self.target_stage = target_stage

        self._pd_component = set.union(*[t.pd_component for t in [self.if_, self.else_]])

    def __call__(self, args):
        assert Transform.pd_stage is not None, ('Stage is not set, use infer_apply, eval_apply '
                                                'or train_apply instead of calling the transform '
                                                'directly.')

        if Transform.pd_stage == self.target_stage:
            return self.if_._pd_call_transform(args)
        return self.else_._pd_call_transform(args)

    def _pd_preprocess_part(self):
        pre = type(self)(
            if_=self.if_.pd_preprocess,
            target_stage=self.target_stage,
            else_=self.else_.pd_preprocess,
        )
        return pre

    def _pd_forward_part(self):
        forward = type(self)(
            if_=self.if_.pd_forward,
            target_stage=self.target_stage,
            else_=self.else_.pd_forward,
        )
        return forward

    def _pd_postprocess_part(self):
        post = type(self)(
            if_=self.if_.pd_postprocess,
            target_stage=self.target_stage,
            else_=self.else_.pd_postprocess,
        )
        return post


class IfInfer(IfInStage):
    """Perform *if_* if called in "infer" stage, else perform *else_*.

    :param if_: Transform for the "infer" stage.
    :param else_: Transform otherwise (defaults to the identity transform).
    """

    def __init__(self, if_: Transform, else_: Optional[Transform] = None):
        super().__init__(if_, 'infer', else_)


class IfEval(IfInStage):
    """Perform *if_* if called in "eval" stage, else perform *else_*.

    :param if_: Transform for the "eval" stage.
    :param else_: Transform otherwise (defaults to the identity transform).
    """

    def __init__(self, if_: Transform, else_: Optional[Transform] = None):
        super().__init__(if_, 'eval', else_)


class IfTrain(IfInStage):
    """Perform *if_* if called in "train" stage, else perform *else_*.

    :param if_: Transform for the "train" stage.
    :param else_: Transform otherwise (defaults to the identity transform).
    """

    def __init__(self, if_: Transform, else_: Optional[Transform] = None):
        super().__init__(if_, 'train', else_)



import torch
from padl.utils import this


def test_this():
    t = this.tolist()
    assert t(torch.tensor([1, 2, 3])) == [1, 2, 3]

    t = this[0]
    assert t([1, 2, 3]) == 1

    t = this[:2]
    assert t([1, 2, 3]) == [1, 2]



import pytest

from padl import transform, group, IfTrain, Batchify, Unbatchify


CONST = 1


@transform
def x(y):
    return CONST + 1


def k(o):
    return o * 7


@transform
def y(y):
    return CONST + k(y)


@transform
def listcomp_a(y):
    return [x + CONST for x in y]


@transform
def listcomp_b(y):
    return [x + y for x in CONST]


@transform
def dictcomp_a(y):
    return {x: x + CONST for x in y}


@transform
def setcomp_a(y):
    return {x + CONST for x in y}


def maketransform():
    @transform
    def z(x):
        return x
    return z


@transform
class MyClassTransform:
    def __init__(self, a, b, c):
        self.a = a
        self.b = b
        self.c = c

    def calculate(self, input_):
        return self.a * self.b + input_

    def __call__(self, input_):
        return self.calculate(input_) + y(input_) + self.c(input_)


def maketransformclass():
    @transform
    class MyClassTransform:
        def __init__(self, a, b, c):
            self.a = a
            self.b = b
            self.c = c

        def calculate(self, input_):
            return self.a * self.b + input_

        def __call__(self, input_):
            return self.calculate(input_) + y(input_) + self.c(input_)

    return MyClassTransform


def makeclasstransform(a, b, c):
    @transform
    class MyClassTransform:
        def __init__(self, a, b, c):
            self.a = a
            self.b = b
            self.c = c

        def calculate(self, input_):
            return self.a * self.b + input_

        def __call__(self, input_):
            return self.calculate(input_) + y(input_) + self.c(input_)

    return MyClassTransform(a, b, c)


def read_dump(name):
    with open(f'tests/material/dumps/{name}.txt') as f:
        return f.read()


def write_dump(dump, name):
    with open(f'tests/material/dumps/{name}.txt', 'w') as f:
        return f.write(dump)


def test_dump_a():
    assert x._pd_dumps() == read_dump('a')


def test_dump_b():
    assert y._pd_dumps() == read_dump('b')


def test_dump_class_a():
    t = MyClassTransform(1, 2, x)
    assert t._pd_dumps() == read_dump('class_a')


lambda_a = transform(lambda x: x)
lambda_b = transform(lambda x: y(x))


def test_lambda_a():
    assert lambda_a._pd_dumps() == read_dump('lambda_a')


def test_lambda_b():
    assert lambda_b._pd_dumps() == read_dump('lambda_b')


def test_nested_dump_a():
    assert maketransform()._pd_dumps() == read_dump('nested_a')


def test_nested_dump_c():
    t = makeclasstransform(1, 2, x)
    assert t._pd_dumps() == read_dump('nested_c')


c_a = x >> y >> x
c_b = x >> y >> x / x + c_a
c_c = (x
    >> Batchify()
    >> y
    >> Unbatchify()
)


def test_compound_dump_a():
    assert c_a._pd_dumps() == read_dump('compound_a')


def test_compound_dump_b():
    assert c_b._pd_dumps() == read_dump('compound_b')


def test_compound_dump_c():
    assert c_c._pd_dumps() == read_dump('compound_c')


g_a = x + group(y + x + x)


def test_grouped_dump_a():
    assert g_a._pd_dumps() == read_dump('grouped_a')


def test_if_train():
    assert IfTrain(x, y)._pd_dumps() == read_dump('iftrain')


def test_listcomp_a():
    assert listcomp_a._pd_dumps() == read_dump('list_comprehension_a')


def test_listcomp_b():
    assert listcomp_b._pd_dumps() == read_dump('list_comprehension_b')


def test_dictcomp_a():
    assert dictcomp_a._pd_dumps() == read_dump('dict_comprehension_a')


def test_setcomp_a():
    assert setcomp_a._pd_dumps() == read_dump('set_comprehension_a')


def test_with_raises():
    with open(__file__) as f:
        x = f.read()

    @transform
    def t(y):
        return x + y

    with pytest.raises(NotImplementedError):
        t._pd_dumps()



from padl import transform


@transform
class MyTransform:
    def __init__(self, x):
        self.x = x

    def __call__(self, y):
        return self.x + y


def make(z):
    return MyTransform(x=z)


def getclass():
    return MyTransform



import pytest

from padl import load, transform
from padl.transforms import ClassTransform, FunctionTransform

import numpy as np
import torch


def test_raise():
    with pytest.raises(ValueError):
        transform(2)


pd_np = transform(np)
pd_torch = transform(torch)


def test_function_a():
    t = pd_np.cos
    assert isinstance(t, FunctionTransform)
    assert t._pd_call == 'pd_np.cos'


def test_function_b():
    t = pd_np.random.rand
    assert isinstance(t, FunctionTransform)
    assert t._pd_call == 'pd_np.random.rand'


def test_class_a():
    t = pd_torch.nn.Linear(10, 10)
    assert isinstance(t, ClassTransform)
    assert t._pd_call == 'pd_torch.nn.Linear(10, 10)'


class TestModuleWrap:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = pd_np.sin
        request.cls.transform_2 = (pd_np.sin >> pd_np.sin)
        transform_temp = pd_np.sin
        request.cls.transform_3 = transform_temp + transform_temp >> pd_np.add
        request.cls.transform_4 = pd_np.cos + pd_np.cos >> pd_np.add

    def test_save_load(self, tmp_path):
        for transform_ in [self.transform_1, self.transform_2, self.transform_3, self.transform_4]:
            transform_.pd_save(tmp_path / 'test.padl', True)
            t_ = load(tmp_path / 'test.padl')
            assert t_.infer_apply(1.3) == transform_.infer_apply(1.3)



"""Utilities for printing. """

from reprlib import repr as shortrepr
from typing import Any, List


def combine_multi_line_strings(strings: List[str]):
    """Combine multi line strings, where every character takes precedence over space.
    (Image space is transparent and superimposing strings on top of one another.)

    :param strings: list of multi-line strings
    """
    strings = [x.split('\n') for x in strings]
    length = max([len(x) for x in strings])
    lines = []
    for i in range(length):
        parts = [x[i] for x in strings if len(x) >= i + 1]
        line = list(' ' * max([len(x) for x in parts]))
        for j in range(len(line)):
            for part in parts:
                if len(part) >= j + 1:
                    if part[j] != ' ':
                        line[j] = part[j]
        lines.append(''.join(line))
    return '\n'.join(lines)


def create_reverse_arrow(start_left: int, finish_right: int, n_initial_rows: int,
                         n_final_rows: int):
    """Create an ascii arrow.

    :param start_left: X-coordinate where the last line arrow ends.
    :param finish_right: X-coor where the first line arrow starts.
    :param n_initial_rows: Y-padding.
    :param n_final_rows: Y-padding.

    Example:

        ____________|
        |
        v
    """
    initial = ''
    for _ in range(n_final_rows - 1):
        initial += ' ' * (start_left + finish_right) + '│' + '\n'

    final = ''
    for _ in range(n_initial_rows - 1):
        final += ' ' * start_left + '│' + '\n'

    underscore_line = list('_' * finish_right)
    if underscore_line:
        underscore_line[0] = ' '
    underscore_line = ''.join(underscore_line)

    output = initial \
        + ' ' * start_left + underscore_line + '│' + ' ' + '\n' \
        + final \
        + ' ' * start_left + '▼'
    output = '\n'.join(output.split('\n')[1:])

    return output


def make_bold(x, disable=False):
    """Make input bold

    :param disable: If *True*, don't apply formatting.
    :param x: string input
    """
    if disable:
        return x
    return f'\33[1m{x}\33[0m'


def make_green(x, disable=False):
    """Make input green

    :param disable: If *True*, don't apply formatting.
    :param x: string input
    """
    if disable:
        return x
    return f'\33[32m{x}\33[0m'


def create_arrow(start_left: int, finish_right: int, n_initial_rows: int, n_final_rows: int) -> str:
    """
    Create an ascii arrow.

    :param start_left: X-coordinate where the last line arrow ends.
    :param finish_right: X-coor where the first line arrow starts.
    :param n_initial_rows: Y-padding.
    :param n_final_rows: Y-padding.

    Example:

        ____________|
        |
        v
    """
    initial = ''
    for _ in range(n_initial_rows - 1):
        initial += ' ' * start_left + '│' + '\n'
    final = ''
    for _ in range(n_final_rows - 1):
        final += ' ' * (start_left + finish_right) + '│' + '\n'
    bend_start = '└' if finish_right else '│'
    bend_end = '┐' if finish_right else ''
    return initial + ' ' * start_left + bend_start \
        + '─' * (finish_right - 1) + bend_end + '\n' \
        + final \
        + ' ' * (start_left + finish_right) + '▼'


def format_argument(value: Any) -> str:
    """Format an argument value for printing."""
    return shortrepr(value)


def visible_len(value: str) -> int:
    """The number of visible characters in a string.

    Use this to get the length of strings that contain escape sequences.
    """
    return int(len(value) - value.count('\x1b') * 4.5 + value.count('\x1b[1'))



"""Module for getting source-code strings of various things (source files, modules, ipython cells).
In addition, the module maintains a cache of source-strings in which the values can be modified.

Use `get_source` to get source-strings given source filenames or ipython cell ids.
`get_module_source` allows to get source of modules.

Both will try to get the source-strings from the `replace_cache`, which can contain modified
versions of the source-strings.

To add a modified source-string to the `replace_cache`, use `put_into_cache`.
"""

import inspect
import linecache
from types import ModuleType
from typing import List


replace_cache = {}


def get_source(filename: str, use_replace_cache: bool = True) -> str:
    """Get source from *filename*.

    The filename is as in the code object gotten from an `inspect.frame`,
    this can be "<ipython input-...>" in which case the source is taken from the ipython cache.

    If *use_replace_cache*, try getting the source from the "replace_cache", which can contain
    explicit replacements of the original source strings.
    """
    if use_replace_cache and filename in replace_cache:
        return replace_cache[filename]
    if filename in linecache.cache:
        # the ipython case
        return ''.join(linecache.cache[filename][2])
    # normal module
    with open(filename) as f:
        return f.read()


def get_module_source(module: ModuleType, use_replace_cache: bool = True) -> str:
    """Get the source code of a module.

    If *use_replace_cache*, try getting the source from the "replace_cache", which can contain
    explicit replacements of the original source strings.
    """
    if use_replace_cache:
        try:
            return replace_cache[module.__filename__]
        except (KeyError, AttributeError):
            pass
    return inspect.getsource(module)


def put_into_cache(key, source: str, repl: str, *loc):
    """Put a string into the "replace_cache".

    Specify the original string, a replacing string and the insertion replacement.

    Example:

        >>> put_into_cache('mykey', 'f(value(x))', 'CONSTANT', 0, 0, 2, 9)
        >>> x = replace_cache['mykey']
        >>> type(x)
        ReplaceString
        >>> x
        "f(CONSTANT)"
        >>> x.original
        "f(value(x))"

    :param key: The key for the cache dict.
    :param source: The source.
    :param repl: The string new inserted / replaced part.
    :param *loc: The location where *repl* is to be inserted (give as *from_line*, *to_line*,
        *from_col*, *to_col*).
    """
    val = ReplaceString(source, repl, *loc)
    try:
        replace_cache[key] = ReplaceStrings([val] + replace_cache[key].rstrings)
    except KeyError:
        replace_cache[key] = ReplaceStrings([val])


def original(string: str) -> str:
    """Get either the original of a `ReplaceString` or a string."""
    return getattr(string, 'original', string)


def cut(string: str, from_line: int, to_line: int, from_col: int, to_col: int) -> str:
    """Cut a string (can be a normal `str`, a `ReplaceString` or a `ReplaceStrings`) and return the
    resulting substring.

    Example:

        Given a *string*,

        "xxxxxxxxxxx
         xxxxxAXXXXXXXXXXXXXXXXXXXX
         XXXXXXXXXXXXXBxxxx
         xxxx"

        , to cut a substring from *A* to *B*, give as *from_line* the line of *A* (1) and *to_line*
        the line of *B* (2). *from_col* determines the position of *A* within the line (5) and
        *to_col* determines the position of *B* (13).
        The result would be:

        "AXXXXXXXXXXXXXXXXXXXX
         XXXXXXXXXXXXXB"

    If used with a `ReplaceString`, the result will be a `ReplaceString` with original and
    replacement at the expected positions.

    :param string: The input string.
    :param from_line: The first line to include.
    :param to_line: The last line to include.
    :param from_col: The first col on *from_line* to include.
    :param to_col: The last col on *to_line* to include.
    :returns: The cut-out string.
    """
    try:
        return string.cut(from_line, to_line, from_col, to_col)
    except AttributeError:
        return _cut_string(string, from_line, to_line, from_col, to_col)


def _ipython_history() -> List[str]:
    """Get the list of commands executed by IPython, ordered from oldest to newest. """
    return [
        replace_cache.get(k, ''.join(lines))
        for k, (_, _, lines, _)
        in linecache.cache.items()
        if k.startswith('<ipython-')
        or 'ipykernel' in k
    ]


class ReplaceString(str):
    """A string with a replaced section.

    Has an attribute `original` which gives the original string (pre-replacing).

    :param string: The original string.
    :param repl: The string to insert at the specified location.
    :param from_line: The first line to replace in.
    :param to_line: The last line to replace in.
    :param from_col: The first col on *from_line* to replace.
    :param to_col: The last col on *to_line* to replace.
    """

    def __new__(cls, string: str, repl: str, from_line: int, to_line: int, from_col: int,
                to_col: int):
        replaced = _replace(string, repl, from_line, to_line, from_col, to_col)
        return super().__new__(cls, replaced)

    def __init__(self, string: str, repl: str, from_line: int, to_line: int, from_col: int,
                 to_col: int):
        self.original = string
        self.from_line = from_line
        self.to_line = to_line
        self.from_col = from_col
        self.to_col = to_col
        self.repl = repl
        super().__init__()

    def cut(self, from_line: int, to_line: int, from_col: int, to_col: int):
        """Cut and return the resulting sub-`ReplaceString`. """
        lines = self.original.split('\n')[from_line: to_line + 1]
        lines[0] = lines[0][from_col:]
        lines[-1] = lines[-1][:to_col]

        new_from_line = self.from_line - from_line
        new_to_line = self.to_line - from_line
        if from_line == self.from_line:
            new_from_col = self.from_col - from_col
        else:
            new_from_col = self.from_col
        if to_line == self.to_line and to_col < self.to_col:
            new_to_col = to_col
        else:
            new_to_col = self.to_col
        if self.to_line == from_line:
            new_to_col -= from_col

        if new_from_line < 0:
            new_from_col = 0
        if new_to_line < 0:
            new_from_col = 0
            new_to_col = 0

        return ReplaceString('\n'.join(lines), self.repl, new_from_line, new_to_line, new_from_col,
                             new_to_col)


class ReplaceStrings(str):
    """A collection of replacestrings with different replacements in an original string.

    :param rstrings: A list of `ReplaceString`s with the same `original` and non-overlapping
        replace locations.
    """

    def __new__(cls, rstrings: List[ReplaceString]):
        assert len(rstrings) > 0, 'Cannot build `ReplaceStrings` with empty list.'
        orig = rstrings[0].original
        assert all(x.original == orig for x in rstrings), \
            'All sub-strings must have the same original'
        # TODO: add assertion to make sure substrings overlap

        rstrings = sorted(rstrings, key=lambda x: (x.from_line, x.from_col), reverse=True)
        replaced = rstrings[0].original
        for rstring in rstrings:
            replaced = _replace(replaced, rstring.repl, rstring.from_line, rstring.to_line,
                                rstring.from_col, rstring.to_col)
        return super().__new__(cls, replaced)

    def __init__(self, rstrings):
        super().__init__()
        self.original = rstrings[0].original
        self.rstrings = rstrings

    def cut(self, from_line, to_line, from_col, to_col):
        """Cut and return the resulting sub-`ReplaceStrings`. """
        return ReplaceStrings([rstr.cut(from_line, to_line, from_col, to_col)
                               for rstr in self.rstrings])


def _cut_string(string: str, from_line: int, to_line: int, from_col: int, to_col: int):
    """Cut a string and return the resulting substring.

    Example:
        Given a *string*,

        "xxxxxxxxxxx
         xxxxxAXXXXXXXXXXXXXXXXXXXX
         XXXXXXXXXXXXXBxxxx
         xxxx"

        , to cut a substring from *A* to *B*, give as *from_line* the line of *A* (1) and *to_line*
        the line of *B* (2). *from_col* determines the position of *A* within the line (5) and
        *to_col* determines the position of *B* (13).
        The result would be:

        "AXXXXXXXXXXXXXXXXXXXX
         XXXXXXXXXXXXXB"

    :param string: The input string.
    :param from_line: The first line to include.
    :param to_line: The last line to include.
    :param from_col: The first col on *from_line* to include.
    :param to_col: The last col on *to_line* to include.
    :returns: The cut-out string.
    """
    lines = string.split('\n')[from_line: to_line + 1]
    if len(lines) == 1:
        return lines[0][from_col:to_col]
    lines[0] = lines[0][from_col:]
    lines[-1] = lines[-1][:to_col]
    return '\n'.join(lines)


def _replace(string, repl, from_line, to_line, from_col, to_col):
    """Replace a substring in *string* with *repl*. """
    if from_line < 0 and to_line < 0:
        return string

    lines = string.split('\n')

    if from_line > len(lines) - 1 and to_line > len(lines) - 1:
        return string

    if from_line < 0:
        from_col = 0
    from_line = max(from_line, 0)
    if to_line > len(lines) - 1:
        to_col = len(lines[-1])
    to_line = min(to_line, len(lines) - 1)

    for i, line in enumerate(lines[:-1]):
        lines[i] = line + '\n'

    before = ''.join(lines[:from_line])
    if from_line == to_line:
        middle = lines[from_line][:from_col] + repl + lines[from_line][to_col:]
        start = end = ''
    else:
        start = lines[from_line][from_col:] + repl
        middle = ''.join(lines[from_line+1:to_line])
        end = lines[to_line][:to_col]
    after = ''.join(lines[to_line+1:])

    return before + start + middle + end + after



"""The Transform class and its fundamental children.

Transforms should be created using the `padl.transform` wrap-function.
"""
import ast
import re
from copy import copy
from collections import Counter, namedtuple, OrderedDict
import contextlib
import inspect
from itertools import chain
from pathlib import Path
from shutil import rmtree
import textwrap
import types
from typing import Callable, Iterable, Iterator, List, Literal, Optional, Set, Tuple, Union
from warnings import warn

import numpy as np
import torch
from torch.utils.data import DataLoader
from tqdm import tqdm

from padl.data import SimpleDataset
from padl.dumptools import var2mod, symfinder, inspector
from padl.dumptools.serialize import Serializer

from padl.dumptools.packagefinder import dump_packages_versions
from padl.exceptions import WrongDeviceError
from padl.print_utils import combine_multi_line_strings, create_reverse_arrow, make_bold, \
    make_green, create_arrow, format_argument, visible_len


class _Notset:
    # pylint: disable=too-few-public-methods
    def __bool__(self):
        return False


_notset = _Notset()


def _isinstance_of_namedtuple(arg):
    """Check if input *arg* is instance of namedtuple"""
    typ = type(arg)
    base = typ.__bases__
    if len(base) != 1 or base[0] != tuple:
        return False
    fields = getattr(typ, '_fields', None)
    if not isinstance(fields, tuple):
        return False
    return all(isinstance(field, str) for field in fields)


def _move_to_device(args, device):
    """Move args to given device

    :param args: args to move to device
    :param device: device to move to
    """
    if isinstance(args, (tuple, list)):
        return tuple([_move_to_device(x, device) for x in args])
    if isinstance(args, torch.Tensor):
        return args.to(device)
    return args


Stage = Literal['infer', 'eval', 'train']
Component = Literal['preprocess', 'forward', 'postprocess']


class Transform:
    """Transform base class.

    :param call_info: A `CallInfo` object containing information about the how the transform was
    created (needed for saving).
    :param pd_name: name of the transform
    """
    pd_stage = None

    def __init__(self, call_info: Optional[inspector.CallInfo] = None,
                 pd_name: Optional[str] = None):
        if call_info is None:
            call_info = inspector.CallInfo()
        self._pd_call_info = call_info
        self._pd_varname = _notset
        self._pd_name = pd_name
        self._pd_component = {'forward'}
        self._pd_device = 'cpu'
        self._pd_layers = None

    @property
    def pd_name(self) -> Optional[str]:
        """The "name" of the transform.

        A transform can have a name. This is optional, but helps when inspecting complex transforms.
        Good transform names indicate what the transform does.

        If a transform does not have an explicitly set name, the name will default to the name of
        the *last variable the transforms was assigned to*.
        """
        if not self._pd_name:
            return self.pd_varname()
        return self._pd_name

    def __rshift__(self, other: "Transform") -> "Compose":
        """Compose with *other*.

        Example:
            t = a >> b >> c
        """
        return Compose([self, other])

    def __add__(self, other: "Transform") -> "Rollout":
        """ Rollout with *other*.

        Example:
            t = a + b + c
        """
        return Rollout([self, other])

    def __truediv__(self, other: "Transform") -> "Parallel":
        """ Parallel with *other*.

        Example:
            t = a / b / c
        """
        return Parallel([self, other])

    def __sub__(self, name: str) -> "Transform":
        """Create a named clone of the transform.

        Example:
            named_t = t - 'rescale image'
        """
        named_copy = copy(self)
        named_copy._pd_name = name
        named_copy._pd_varname = _notset
        return named_copy

    def __invert__(self) -> "Map":
        """Map.

        Example:
            t = ~a
        """
        return Map(self)

    def pd_pre_save(self, path: Path, i: int):
        """Method that is called on each transform before saving.

        This normally does nothing. Override to implement custom serialization.

        :param path: The save-folder path.
        :param i: Unique transform index, can be used to construct filenames.
        """

    def pd_post_load(self, path: Path, i: int):
        """Method that is called on each transform after loading.

        This normally does nothing. Override to implement custom serialization.

        :param path: The load path.
        :param i: Unique transform index, can be used to construct filenames.
        """

    def pd_save(self, path: Union[Path, str], force_overwrite: bool = False):
        """Save the transform to a folder at *path*.

        The folder's name should end with '.padl'. If no extension is given, it will be added
        automatically.

        If the folder exist, call with *force_overwrite* = `True` to overwrite. Otherwise, this
        will raise a FileExistsError.
        """
        path = Path(path)
        if path.suffix == '':
            path = path.parent / (path.name + '.padl')

        if path.exists():
            if not force_overwrite:
                raise FileExistsError(f'{path} exists, call with *force_overwrite* to overwrite.')
            rmtree(path)

        path.mkdir()

        for i, subtrans in enumerate(self._pd_all_transforms()):
            subtrans.pd_pre_save(path, i)
        code, versions = self._pd_dumps(True, path=path)

        with open(path / 'transform.py', 'w') as f:
            f.write(code)
        with open(path / 'versions.txt', 'w') as f:
            f.write(versions)

    def _pd_codegraph_startnode(self, name: str) -> var2mod.CodeNode:
        """Build the start-code-node - the node with the source needed to create *self* as "name".
        (in the scope where *self* was originally created). """
        start_source = f'{name or "_pd_dummy"} = {self._pd_evaluable_repr()}'
        start_node = ast.parse(start_source).body[0]
        start_globals = {
            (var, self._pd_call_info.scope)  # this should be the current scope ...?
            for var in var2mod.find_globals(start_node)
        }
        return var2mod.CodeNode(
            source=start_source,
            ast_node=start_node,
            globals_=start_globals
        )

    @property
    def _pd_closurevars(self) -> Tuple[dict, dict]:
        """Return the closurevars (globals and nonlocals) the transform depends on. """
        return {}, {}

    def _pd_build_codegraph(self, graph: Optional[dict] = None,
                            scopemap: Optional[dict] = None,
                            name: Optional[str] = None,
                            scope: Optional[symfinder.Scope] = None) -> Tuple[dict, dict]:
        if graph is None:
            graph = {}
        if scopemap is None:
            scopemap = {}

        if scope is None:
            scope = self._pd_call_info.scope

        given_name = name

        try:
            if self._pd_call == name:
                name = None
        except AttributeError:
            pass

        # build the start node ->
        start = self._pd_codegraph_startnode(name)
        # <-

        globals_dict, nonlocals_dict = self._pd_closurevars
        all_vars_dict = {**globals_dict, **nonlocals_dict}

        # find dependencies
        todo = {*start.globals_}
        while todo and (next_ := todo.pop()):
            # we know this already - go on
            if next_ in scopemap:
                continue

            next_var, next_scope = next_

            if next_var.startswith('PADL_VALUE'):
                continue

            # see if the object itself knows how to generate its codegraph
            try:
                if len(next_scope) > 0:
                    next_obj = all_vars_dict[next_var]
                else:
                    next_obj = globals_dict[next_var]
                next_obj._pd_build_codegraph(graph, scopemap, next_var)
            except (KeyError, AttributeError):
                pass
            else:
                continue

            # find how next_var came into being
            (source, node), scope_of_next_var = symfinder.find_in_scope(next_var, next_scope)
            scopemap[next_var, next_scope] = scope_of_next_var

            # find dependencies
            globals_ = {
                (var, scope_of_next_var)
                for var in var2mod.find_globals(node)
            }
            graph[next_var, scope_of_next_var] = var2mod.CodeNode(source=source, globals_=globals_,
                                                                  ast_node=node)
            todo.update(globals_)
        # find dependencies done

        if name is not None:
            assert scope is not None
            graph[name, scope] = start

        if given_name is not None:
            scopemap[given_name, scope] = scope

        return graph, scopemap

    def _pd_evaluable_repr(self, indent: int = 0) -> str:
        """Return a string that if evaluated *in the same scope where the transform was created*
        creates the transform. """
        result = self._pd_evaluable_repr_inner(indent)
        if self._pd_name is not None:
            return f"({result} - '{self._pd_name}')"
        return result

    def _pd_evaluable_repr_inner(self, indent: int = 0) -> str:
        # pylint: disable=unused-argument,no-self-use
        """Return a string that if evaluated *in the same scope where the transform was created*
        creates the transform. """
        raise NotImplementedError

    def _pd_all_transforms(self, result: Optional[list] = None) -> list:
        """Return a list of all transforms needed for executing the transform.

        This includes the transform itself, the subtransforms of a compount transform or
        transforms a function-transform depends on as a global. """
        if result is None:
            result = []
        if self in result:
            return result
        result.append(self)
        for transform in self._pd_direct_subtransforms:
            # pylint: disable=protected-access
            transform._pd_all_transforms(result)
        return result

    @property
    def _pd_direct_subtransforms(self) -> Iterator['Transform']:
        """Iterator over the direct subtransforms. """
        # pylint: disable=no-self-use
        raise NotImplementedError

    def _pd_dumps(self, return_versions: bool = False,
                  path: Optional[Path] = None) -> Union[str, Tuple[str, str]]:
        """Dump the transform as python code.

        :param return_versions: If *True* return a tuple of the code and a file listing
            dependencies and their versions.
        :param path: Optional path to save at, might be required for serializer code snippets.
        """
        scope = symfinder.Scope.toplevel(inspector.caller_module())
        graph, scopemap = self._pd_build_codegraph(name='_pd_main', scope=scope)
        Serializer.save_all(graph, scopemap, path)
        unscoped = var2mod.unscope_graph(graph, scopemap)
        code = var2mod.dumps_graph(unscoped)
        if return_versions:
            versions = dump_packages_versions(node.ast_node for node in graph.values())
            return code, versions
        return code

    def __repr__(self):
        return self._pd_shortrepr(formatting=False)

    def _repr_pretty_(self, p, cycle) -> str:
        # pylint: disable=invalid-name
        title = self._pd_title()
        if self.pd_name is not None and self.pd_name != title:
            title = make_bold(title) + f' - "{self.pd_name}"'
        else:
            title = make_bold(title)
        top_message = title + ':' + '\n\n'
        bottom_message = textwrap.indent(self._pd_longrepr(), '   ')
        p.text(top_message + bottom_message if not cycle else '...')

    def _pd_longrepr(self, formatting=True) -> str:
        """A lone string representation of the transform."""
        raise NotImplementedError

    def _pd_parensrepr(self, formatting=True) -> str:
        short = self._pd_shortrepr(formatting)
        if len(short) < 50:
            if len(getattr(self, 'transforms', [])) > 1:
                short = f"{make_green('[', not formatting)}{short}{make_green(']', not formatting)}"
            return short
        return self._pd_tinyrepr(formatting)


    def _pd_shortrepr(self, formatting=True) -> str:
        """A short string representation of the transform."""
        return self._pd_title()

    def _pd_tinyrepr(self, formatting=True) -> str:
        """A tiny string representation of the transform."""
        return self.pd_name or f'<anonymous {self.__class__.__name__}>'

    def _pd_title(self) -> str:
        """A title for the transform."""
        return self._pd_tinyrepr()

    def _pd_find_varname(self, scopedict: dict) -> Optional[str]:
        """Find the name of the variable name the transform was last assigned to.

        :return: A string with the variable name or *None* if the transform has not been assigned
            to any variable.
        """
        try:
            return [
                k for k, v in scopedict.items()
                if v is self and not k.startswith('_')
            ][0]
        except IndexError:
            return None

    def pd_varname(self) -> Optional[str]:
        """The name of the variable name the transform was last assigned to.

        Example:

        >>> foo = MyTransform()
        >>> foo._pd_varname
        "foo"

        :return: A string with the variable name or *None* if the transform has not been assigned
            to any variable.
        """
        if self._pd_varname is _notset:
            module = inspector.caller_module()
            self._pd_varname = self._pd_find_varname(module.__dict__)
        return self._pd_varname

    def _pd_forward_device_check(self) -> bool:
        """Check if all transform in forward are in correct device

        All transforms in forward need to be in same device as specified for
        the whole CompoundTransform.
        """
        for layer in self.pd_forward.pd_layers:
            for parameters in layer.parameters():
                parameter_device = parameters.device.type
                if ':' in self.pd_device and 'cuda' in parameter_device:
                    parameter_device += f':{parameters.device.index}'
                if parameter_device != self.pd_device:
                    raise WrongDeviceError(self, layer)
        return True

    def _pd_unpack_argument(self, arg) -> bool:
        """Return *True* if to arguments should be unpacked, else *False*"""
        signature_count = 0
        if not isinstance(arg, (list, tuple)):
            return False

        if hasattr(self, '_pd_number_of_inputs') and self._pd_number_of_inputs is not None:
            return self._pd_number_of_inputs > 1

        try:
            parameters = self._pd_get_signature().values()
        except ValueError:
            return False
        for param in parameters:
            if param.kind in (
                    param.POSITIONAL_OR_KEYWORD,
                    param.POSITIONAL_ONLY):
                signature_count += 1
            if param.kind == param.VAR_POSITIONAL:
                return True
        if signature_count > 1:
            return True
        return False

    def _pd_call_transform(self, arg, stage: Optional[Stage] = None):
        """Call the transform, with possibility to pass multiple arguments.

        :param stage: The stage ("infer", "eval", "train") to perform the call with.
        :return: Whatever the transform returns.
        """

        if stage in ('eval', 'infer'):
            torch_context = torch.no_grad()
        else:
            torch_context = contextlib.suppress()

        with self.pd_set_stage(stage), torch_context:
            if self._pd_unpack_argument(arg):
                return self(*arg)
            return self(arg)

    def _pd_get_signature(self):
        """Get the signature of the transform. """
        return inspect.signature(self).parameters

    def _pd_itercall(self, args, stage: Stage, loader_kwargs: Optional[dict] = None,
                     verbose: bool = False, flatten: bool = False) -> Iterator:
        """Create a data loader and run preprocessing, forward, and postprocessing steps.

        :param args: Arguments to call with.
        :param stage: Stage to call in ("eval", "train" or "infer")
        :param loader_kwargs: Data loader keyword arguments.
        :param verbose: If *True*, print progress bar.
        :param flatten: If *True*, flatten the output.

        :return: A generator that allows iterating over the output.
        """
        assert stage in ('eval', 'train'), '_pd_itercall can only be used with stage eval or train'

        self._pd_forward_device_check()

        preprocess = self.pd_preprocess
        forward = self.pd_forward
        post = self.pd_postprocess

        use_preprocess = not isinstance(preprocess, Identity)
        use_forward = not isinstance(forward, Identity)
        use_post = not isinstance(post, Identity)

        if use_preprocess:
            data = SimpleDataset(
                args,
                lambda *args: self.pd_preprocess._pd_call_transform(*args, stage),
            )
            if loader_kwargs is None:
                loader_kwargs = {}
            loader = self._pd_get_loader(sequence=data, loader_kwargs=loader_kwargs)
        else:
            loader = args

        pbar = None
        if verbose:
            if flatten:
                pbar = tqdm(total=len(args))
            else:
                loader = tqdm(loader, total=len(loader))

        for batch in loader:
            batch = _move_to_device(batch, self.pd_device)

            if use_forward:
                output = forward._pd_call_transform(batch, stage)
            else:
                output = batch

            if use_post:
                output = post._pd_call_transform(output, stage)

            if flatten:
                if verbose:
                    pbar.update()
                if not use_post:
                    output = Unbatchify(cpu=False)(batch)
                if hasattr(self, '_pd_output_format'):
                    yield from self._pd_output_format(*output)
                else:
                    yield from output
                continue
            if hasattr(self, '_pd_output_format'):
                yield self._pd_output_format(*output)
            else:
                yield output

    @property
    def pd_device(self) -> str:
        """Return the device ("cpu" / "cuda") the transform is on."""
        return self._pd_device

    @property
    def pd_component(self) -> Set[Component]:
        """Return the component (preprocess, forward or postprocess)."""
        return self._pd_component

    def _pd_preprocess_part(self) -> "Transform":
        return Identity()

    @property
    def pd_preprocess(self) -> "Transform":
        """The preprocessing part of the transform. The device must be propagated from self."""
        if {'preprocess'} == self.pd_component:
            return self
        pre = self._pd_preprocess_part()
        pre.pd_to(self.pd_device)
        return pre

    def _pd_forward_part(self) -> "Transform":
        return Identity()

    @property
    def pd_forward(self) -> "Transform":
        """The forward part of the transform (that what's typically done on the GPU).
        The device must be propagated from self."""
        if {'forward'} == self.pd_component:
            return self
        forward = self._pd_forward_part()
        forward.pd_to(self.pd_device)
        return forward

    def _pd_postprocess_part(self) -> "Transform":
        return Identity()

    @property
    def pd_postprocess(self) -> "Transform":
        """The postprocessing part of the transform. The device must be propagated from self."""
        if {'postprocess'} == self.pd_component:
            return self
        post = self._pd_postprocess_part()
        post.pd_to(self.pd_device)
        return post

    def pd_to(self, device: str) -> "Transform":
        """Set the transform's device to *device*.

        :param device: Device to set the transform to {'cpu', 'cuda', 'cuda:N'}.
        """
        self._pd_device = device
        for layer in self.pd_layers:
            layer.to(device)
        return self

    @property
    def pd_layers(self) -> List[torch.nn.Module]:
        """Get a list with all pytorch layers in the transform (including layers in sub-transforms).
        """
        layers = []
        for subtrans in self._pd_all_transforms():
            if isinstance(subtrans, torch.nn.Module):
                layers.append(subtrans)
        return layers

    def pd_parameters(self) -> Iterator:
        """Iterate over all (pytorch-) parameters in all layers contained in the transform. """
        for layer in self.pd_layers:
            yield from layer.parameters()

    @contextlib.contextmanager
    def pd_set_stage(self, stage: Optional[str] = None):
        """Set of stage of Transform

        :param stage: stage ('train', 'eval', 'infer')
        """
        assert stage in ('train', 'eval', 'infer', None)

        if stage is None:
            yield
            return

        layers = self.pd_layers
        training_before = [layer.training for layer in layers]
        try:
            for layer in layers:
                if stage == 'train':
                    layer.train()
                else:
                    layer.eval()
            Transform.pd_stage = stage
            yield
        finally:
            for i, training in enumerate(training_before):
                layer = layers[i]
                if training:
                    layer.train()
                else:
                    layer.eval()
            Transform.pd_stage = None

    @staticmethod
    def _pd_get_loader(sequence, loader_kwargs=None) -> DataLoader:
        """Get a pytorch data loader.

        :param sequence: A sequence of datapoints.
        :param loader_kwargs: Keyword arguments passed to the data loader (see the pytorch
            `DataLoader` documentation for details).
        """
        return DataLoader(
            sequence,
            worker_init_fn=lambda _: np.random.seed(),
            **loader_kwargs
        )

    def infer_apply(self, inputs):
        """Call transform within the infer context.

        This expects a single argument and returns a single output.

        :param inputs: The input.
        """
        inputs = self.pd_preprocess._pd_call_transform(inputs, stage='infer')
        inputs = _move_to_device(inputs, self.pd_device)
        inputs = self.pd_forward._pd_call_transform(inputs, stage='infer')
        inputs = self.pd_postprocess._pd_call_transform(inputs, stage='infer')

        return inputs

    def eval_apply(self, inputs: Iterable,
                   verbose: bool = False, flatten: bool = False, **kwargs):
        """Call transform within the eval context.

        This will use multiprocessing for the preprocessing part via `DataLoader` and turn
        of gradients for the forward part.

        It expects an iterable input and returns a generator.

        :param inputs: The arguments - an iterable (e.g. list) of inputs.
        :param kwargs: Keyword arguments to be passed on to the dataloader. These can be
            any that a `torch.data.utils.DataLoader` accepts.
        :param verbose: If *True*, print progress bar.
        :param flatten: If *True*, flatten the output.
        """
        return self._pd_itercall(inputs, 'eval', loader_kwargs=kwargs,
                                 verbose=verbose, flatten=flatten)

    def train_apply(self, inputs: Iterable,
                    verbose: bool = False, flatten: bool = False, **kwargs):
        """Call transform within the train context.

        This will use multiprocessing for the preprocessing part via `DataLoader` and turn
        on gradients for the forward part.

        It expects an iterable input and returns a generator.

        :param inputs: The arguments - an iterable (e.g. list) of inputs.
        :param kwargs: Keyword arguments to be passed on to the dataloader. These can be
            any that a `torch.data.utils.DataLoader` accepts.
        :param verbose: If *True*, print progress bar.
        :param flatten: If *True*, flatten the output.
        """
        return self._pd_itercall(inputs, 'train', loader_kwargs=kwargs,
                                 verbose=verbose, flatten=flatten)


class AtomicTransform(Transform):
    """Base class for "atomic" transforms (transforms that are not made by combining
    other transforms - in contrast to `CompoundTransform`s).

    Examples of `AtomicTransform`s are `ClassTransform`s and `FunctionTransform`s.

    :param call: The transform's call string.
    :param call_info: A `CallInfo` object containing information about the how the transform was
        created (needed for saving).
    :param pd_name: The transform's name.
    """

    def __init__(self, call: str, call_info: Optional[inspector.CallInfo] = None,
                 pd_name: Optional[str] = None):
        super().__init__(call_info, pd_name)
        self._pd_call = call

    def _pd_evaluable_repr_inner(self, indent: int = 0) -> str:
        return self._pd_call

    def _pd_title(self) -> str:
        return self._pd_call

    @property
    def _pd_direct_subtransforms(self) -> Iterator[Transform]:
        # pylint: disable=no-self-use
        globals_dict, nonlocals_dict = self._pd_closurevars
        for v in chain(self.__dict__.values(), globals_dict.values(), nonlocals_dict.values()):
            if isinstance(v, Transform):
                yield v


class FunctionTransform(AtomicTransform):
    """A transform that wraps a *function*.

    Do not use this directly - rather, wrap a function using `padl.transform`,

    as a decorator::

        @transform
        def f(x):
            ...

    inline::

        t = transform(f)

    or with a lambda function::

        t = transform(lambda x: x + 1)

    :param function: The wrapped function.
    :param call_info: A `CallInfo` object containing information about the how the transform was
        created (needed for saving).
    :param pd_name: name of the transform
    :param call: The call string (defaults to the function's name).
    :param source: The source code (optional).
    """

    def __init__(self, function: Callable, call_info: inspector.CallInfo,
                 pd_name: Optional[str] = None, call: Optional[str] = None,
                 source: Optional[str] = None):
        if call is None:
            call = function.__name__
        super().__init__(call=call, call_info=call_info, pd_name=pd_name)
        self.function = function
        self._pd_number_of_inputs = None
        self._source = source

    @property
    def source(self) -> str:
        """The source of the wrapped function. """
        if self._source is not None:
            return self._source
        body_msg = inspect.getsource(self.function)
        body_msg = ''.join(re.split('(def )', body_msg, 1)[1:])
        return body_msg

    def _pd_get_signature(self) -> List[str]:
        if self._pd_number_of_inputs is None:
            return inspect.signature(self).parameters
        return [f'arg_{i}' for i in range(self._pd_number_of_inputs)]

    def _pd_longrepr(self, formatting=True) -> str:
        try:
            return '\n'.join(self.source.split('\n')[:30])
        except TypeError:
            return self._pd_call

    def _pd_shortrepr(self, formatting=True) -> str:
        if len(self._pd_longrepr().split('\n', 1)) == 1:
            return self._pd_longrepr(formatting)
        return super()._pd_shortrepr(formatting)

    def _pd_title(self) -> str:
        return self.function.__name__

    @property
    def _pd_closurevars(self) -> inspect.ClosureVars:
        try:
            closurevars = inspect.getclosurevars(self.function)
        except TypeError as exc:
            warn(f'Could not get closurevars ({exc}). This is usually fine as closurevars are only '
                 'needed for user defined transforms.',
                 RuntimeWarning)
            return {}, {}
        return closurevars.globals, closurevars.nonlocals

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)


class ClassTransform(AtomicTransform):
    """Class Transform.

    Do not use this directly, instead, use the `transform` decorator to wrap a class.

    :param pd_name: name of the transform
    :param ignore_scope: Don't try to determine the scope (use the toplevel scope instead).
    :param arguments: ordered dictionary of initialization arguments to be used in printing
    """

    def __init__(self, pd_name: str = None, ignore_scope: bool = False,
                 arguments: Optional[OrderedDict] = None):
        caller_frameinfo = inspector.non_init_caller_frameinfo()
        call_info = inspector.CallInfo(caller_frameinfo, ignore_scope=ignore_scope)
        call = inspector.get_segment_from_frame(caller_frameinfo.frame, 'call')
        call = re.sub(r'\n\s*', ' ', call)
        self._pd_arguments = arguments
        AtomicTransform.__init__(
            self,
            call=call,
            call_info=call_info,
            pd_name=pd_name
        )

    @property
    def source(self) -> str:
        """The class source code. """
        (body_msg, _), _ = symfinder.find_in_scope(self.__class__.__name__,
                                                   self._pd_call_info.scope)
        try:
            return 'class ' + body_msg.split('class ', 1)[1]
        except IndexError:
            return body_msg

    def _formatted_args(self) -> str:
        """Format the object's init arguments for printing. """
        if self._pd_arguments is None:
            return '-?-'

        args_list = []
        for key, value in self._pd_arguments.items():
            if key == 'args':
                args_list += [f'{format_argument(val)}' for val in value]
            elif key == 'kwargs':
                args_list += [f'{subkey}={format_argument(val)}' for subkey, val in value.items()]
            else:
                args_list.append(f'{key}={format_argument(value)}')
        return ', '.join(args_list)

    def _pd_longrepr(self) -> str:
        try:
            return '\n'.join(self.source.split('\n')[:30])
        except symfinder.NameNotFound:
            return self._pd_call

    def _pd_title(self) -> str:
        title = type(self).__name__
        return title + '(' + self._formatted_args() + ')'


class TorchModuleTransform(ClassTransform):
    """Transform class for use with `torch.nn.Module`."""

    def _pd_get_signature(self):
        return inspect.signature(self.forward).parameters

    def pd_pre_save(self, path: Path, i: int):
        """Dump the model's parameters to a save-folder.

        :param path: The save-folder path.
        :param i: Unique transform index, used to construct filenames.
        """
        path = Path(path)
        checkpoint_path = path / f'{path.stem}_{i}.pt'
        print('saving torch module to', checkpoint_path)
        torch.save(self.state_dict(), checkpoint_path)

    def pd_post_load(self, path, i):
        """Load the model's parameters form a save-folder.

        :param path: The save-folder path.
        :param i: Unique transform index, used to construct filenames.
        """
        path = Path(path)
        checkpoint_path = path / f'{path.stem}_{i}.pt'
        print('loading torch module from', checkpoint_path)
        self.load_state_dict(torch.load(checkpoint_path))

    def _pd_longrepr(self) -> str:
        return torch.nn.Module.__repr__(self)


class Map(Transform):
    """Apply one transform to each element of a list.

    >>> Map(t)([x1, x2, x3]) == [t(x1), t(x2), t(x3)]
    True

    :param transform: Transform to be applied to a list of inputs.
    :param call_info: A `CallInfo` object containing information about the how the transform was
    created (needed for saving).
    :param pd_name: name of the transform
    """

    def __init__(self, transform: Transform, call_info: Optional[inspector.CallInfo] = None,
                 pd_name: Optional[str] = None):
        super().__init__(call_info, pd_name)

        self.transform = transform
        self._pd_component = transform.pd_component

    def __call__(self, args: Iterable):
        """
        :param args: Args list to call transforms with
        """
        return [self.transform._pd_call_transform(arg) for arg in args]

    def _pd_longrepr(self) -> str:
        return '~ ' + self.transform._pd_shortrepr()

    @property
    def _pd_direct_subtransforms(self) -> Iterator[Transform]:
        yield self.transform

    def _pd_evaluable_repr_inner(self, indent: int = 0) -> str:
        varname = self.transform.pd_varname()
        if varname:
            return f'~{varname}'
        return f'~{self.transform._pd_evaluable_repr(indent)}'

    def _pd_build_codegraph(self, graph=None, scopemap=None, name=None, scope=None):
        if graph is None:
            graph = {}
        if scopemap is None:
            scopemap = {}

        start = self._pd_codegraph_startnode(name)

        if name is not None:
            assert scope is not None
            graph[name, scope] = start
            scopemap[name, scope] = scope

        varname = self.transform.pd_varname()
        self.transform._pd_build_codegraph(graph, scopemap, varname,  self._pd_call_info.scope)
        return graph, scopemap

    def _pd_preprocess_part(self) -> Transform:
        t_pre = self.transform.pd_preprocess
        if isinstance(t_pre, Identity):
            pre = Identity()
        else:
            pre = Map(transform=t_pre, call_info=self._pd_call_info)
        return pre

    def _pd_postprocess_part(self) -> Transform:
        t_post = self.transform.pd_postprocess
        if isinstance(t_post, Identity):
            post = Identity()
        else:
            post = Map(transform=t_post, call_info=self._pd_call_info)
        return post

    def _pd_forward_part(self) -> Transform:
        t_for = self.transform.pd_forward
        if isinstance(t_for, Identity):
            forward = Identity()
        else:
            forward = Map(transform=t_for, call_info=self._pd_call_info)
        return forward


class CompoundTransform(Transform):
    """Abstract base class for compound-transforms (transforms combining other transforms).

    :param transforms: list of transforms
    :param call_info: A `CallInfo` object containing information about the how the transform was
    created (needed for saving).
    :param pd_name: name of CompoundTransform
    :param pd_group: If *True*, do not flatten this when used as child transform in a
        `CompoundTransform`.
    """
    op = NotImplemented
    display_op = NotImplemented

    def __init__(self, transforms, call_info=None, pd_name=None, pd_group=False):
        super().__init__(call_info, pd_name)

        self._pd_group = True if pd_name is not None else pd_group

        transforms = self._flatten_list(transforms)
        self.transforms: List[Transform] = transforms

        self._pd_component_list = [t.pd_component for t in self.transforms]
        try:
            self._pd_component = set.union(*self._pd_component_list)
        except (AttributeError, TypeError):
            self._pd_component = None

    def __sub__(self, name: str) -> "Transform":
        """Create a named clone of the transform.

        Example:
            named_t = t - 'rescale image'
        """
        named_copy = copy(self)
        named_copy._pd_name = name
        named_copy._pd_group = True
        named_copy._pd_varname = _notset
        return named_copy

    def __getitem__(self, item: Union[int, slice, str]) -> Transform:
        """Get item

        If int, gets item'th transform in this CompoundTransform.
        If slice, gets sliced transform of same type
        If str, gets first transform with name item

        :param item: Should be of type {int, slice, str}
        """
        if isinstance(item, int):
            return self.transforms[item]
        if isinstance(item, slice):
            transform_ = type(self)(self.transforms[item])
            transform_.pd_to(self.pd_device)
            return transform_
        if isinstance(item, str):
            for transform_ in self.transforms:
                if transform_.pd_name == item:
                    return transform_
            raise ValueError(f"{item}: Transform with pd_name '{item}' not found")
        raise TypeError('Unknown type for get item: expected type {int, slice, str}')

    def _pd_evaluable_repr_inner(self, indent=0):
        sub_reprs = [
            x.pd_varname() or x._pd_evaluable_repr(indent + 4)
            for x in self.transforms
        ]
        result = (
            '(\n    ' + ' ' * indent
            + ('\n' + ' ' * indent + f'    {self.op} ').join(sub_reprs)
            + '\n' + ' ' * indent + ')'
        )
        if self._pd_group:
            result = 'padl.group' + result
        return result

    def _pd_build_codegraph(self, graph=None, scopemap=None, name=None, scope=None):
        if graph is None:
            graph = {}
        if scopemap is None:
            scopemap = {}

        start = self._pd_codegraph_startnode(name)

        if self._pd_group and 'padl' not in graph:
            emptyscope = symfinder.Scope.empty()
            graph['padl', emptyscope] = var2mod.CodeNode.from_source('import padl', emptyscope)
            scopemap['padl', self._pd_call_info.scope] = emptyscope

        if name is not None:
            assert scope is not None
            graph[name, scope] = start
            scopemap[name, scope] = scope

        for transform in self.transforms:
            varname = transform.pd_varname()
            transform._pd_build_codegraph(graph, scopemap, varname,
                                          self._pd_call_info.scope)
        return graph, scopemap

    def _pd_longrepr(self, formatting=True):
        between = f'\n{make_green(self.display_op, not formatting)}  \n'
        rows = [make_bold(f'{i}: ', not formatting) + t._pd_shortrepr(formatting)
                for i, t in enumerate(self.transforms)]
        return between.join(rows) + '\n'

    def _pd_shortrepr(self, formatting=True):
        def subrepr(transform):
            short = transform._pd_shortrepr(formatting)
            if len(short) < 20:
                return short
            return transform._pd_tinyrepr(formatting)

        result = f' {make_green(self.op, not formatting)} ' \
            .join(subrepr(t) for t in self.transforms)
        if self._pd_group:
            result = f'group({result})'
        return result

    def _pd_tinyrepr(self, formatting=True) -> str:
        rep = f'..{make_green(self.op, not formatting)}..'
        if self.pd_name:
            rep = f'{self.pd_name}: {rep}'
        return f'{make_green("[", not formatting)}{rep}{make_green("]", not formatting)}'

    def _pd_title(self):
        return self.__class__.__name__

    def pd_to(self, device: str):
        """Set the transform's device to *device*

        :param device: device on which to send {'cpu', cuda', 'cuda:N'}
        """
        self._pd_device = device
        for transform_ in self.transforms:
            transform_.pd_to(device)
        return self

    def _pd_forward_device_check(self):
        """Check all transform in forward are in correct device

        All transforms in forward need to be in same device as specified for
        the whole CompoundTransform.

        :return: Bool
        """
        return_val = True

        if isinstance(self.pd_forward, type(self)):
            for transform_ in self.pd_forward.transforms:
                if self.pd_device != transform_.pd_device:
                    raise WrongDeviceError(self, transform_)
                return_val = transform_._pd_forward_device_check()
            return return_val

        if self.pd_device != self.pd_forward.pd_device:
            raise WrongDeviceError(self, self.pd_forward)

        return self.pd_forward._pd_forward_device_check()

    @classmethod
    def _flatten_list(cls, transform_list: List[Transform]):
        """Flatten *list_* such that members of *cls* are not nested.

        :param transform_list: List of transforms.
        """
        list_flat = []

        for transform in transform_list:
            if isinstance(transform, cls):
                if transform._pd_group:
                    list_flat.append(transform)
                else:
                    list_flat += transform.transforms
            else:
                list_flat.append(transform)

        return list_flat

    @property
    def _pd_direct_subtransforms(self):
        yield from self.transforms

    def grouped(self):
        """Return a grouped version of *self*. """
        transform_ = type(self)(self.transforms, self._pd_call_info, pd_name=self.pd_name,
                                pd_group=True)
        transform_.pd_to(self.pd_device)
        return transform_

    @staticmethod
    def _pd_get_keys(transforms):
        """Get deduplicated keys from list of transforms

        Names are updated as below.
        [None, None, 'a', 'a', 'b', None] -> ['out_0', 'out_1', 'a_0', 'a_1', 'b', 'out_5']

        :param transforms: list of transforms
        :return: list of keys
        """
        names = []
        for ind, transform_ in enumerate(transforms):
            if not transform_.pd_name:
                name = 'out_'+str(ind)
            else:
                name = transform_.pd_name
            names.append(name)

        counter = Counter(names)
        updated_counter = Counter()
        deduped_keys = []

        for name in names:
            new_name = name + '_' + str(updated_counter[name]) if counter[name] > 1 else name
            updated_counter.update({name: 1})
            deduped_keys.append(new_name)
        return deduped_keys


class Compose(CompoundTransform):
    """Apply series of transforms on input.

    Compose([t1, t2, t3])(x) = t3(t1(t2(x)))

    :param transforms: List of transforms to compose.
    :param call_info: A `CallInfo` object containing information about the how the transform was
        created (needed for saving).
    :param pd_name: name of the Compose transform
    :param pd_group: If *True*, do not flatten this when used as child transform in a
        `CompoundTransform`.
    :return: output from series of transforms
    """
    op = '>>'
    display_op = '>>'

    def __init__(self, transforms: Iterable[Transform], call_info: inspector.CallInfo = None,
                 pd_name: Optional[str] = None, pd_group: bool = False):
        super().__init__(transforms, call_info=call_info, pd_name=pd_name, pd_group=pd_group)

        preprocess_end = 0
        postprocess_start = len(self.transforms)
        set_postprocess = True
        for i, transform_ in enumerate(self.transforms):
            if 'preprocess' in transform_.pd_component:
                preprocess_end = i
            if 'postprocess' in transform_.pd_component and set_postprocess:
                postprocess_start = i
                set_postprocess = False
        for i in range(preprocess_end):
            self._pd_component_list[i] = {'preprocess'}
        for i in range(postprocess_start+1, len(self.transforms)):
            self._pd_component_list[i] = {'postprocess'}

    def _pd_classify_nodetype(self, i, t, t_m1, cw, cw_m1):
        if i > 0 and isinstance(t, Parallel) and len(cw) == len(cw_m1):
            type_ = 'multi_2_multi'

        elif i > 0 and cw == 1 and cw_m1 > 1:
            type_ = 'multi_2_single'

        elif cw == 1 or isinstance(t, Compose):
            type_ = 'single_2_single'

        else:
            type_ = 'single_2_multi'

        return type_

    def _pd_longrepr(self, formatting=True) -> str:  # TODO: make it respect the formatting
        """Create a detailed formatted representation of the transform. For multi-line inputs
        the lines are connected with arrows indicating data flow.
        """
        # pad the components of rows which are shorter than other parts in same column
        rows = [
            [s._pd_parensrepr() for s in t.transforms] if hasattr(t, 'transforms')
            else [t._pd_shortrepr()]
            for t in self.transforms
        ]

        children_widths = [[visible_len(x) for x in row] for row in rows]
        # get maximum widths in "columns"
        children_widths_matrix = np.zeros((len(self.transforms),
                                           max([len(x) for x in children_widths])))
        for i, cw in enumerate(children_widths):
            children_widths_matrix[i, :len(cw)] = cw
        max_widths = np.max(children_widths_matrix, 0)

        for i, r in enumerate(rows):
            for j in range(len(rows[i])):
                if len(rows[i][j]) < max_widths[j]:
                    rows[i][j] += ' ' * (int(max_widths[j]) - len(rows[i][j]))

        for i, r in enumerate(rows):
            if len(r) > 1:
                rows[i] = f' {make_green(self.transforms[i].display_op)} '.join(r)
            else:
                rows[i] = r[0]
        output = []
        # iterate through rows and create arrows depending on numbers of components
        for i, (r, t) in enumerate(zip(rows, self.transforms)):
            widths = [0]
            subarrows = []

            type_ = self._pd_classify_nodetype(i, t, self.transforms[i - 1],
                                               children_widths[i], children_widths[i - 1])

            # if subsequent rows have the same number of "children" transforms
            if type_ == 'multi_2_multi':
                for j, w in enumerate(children_widths[i]):
                    subarrows.append(create_arrow(sum(widths) - j + j * 4, 0, 0, 0))
                    widths.append(int(max_widths[j]))

            # if previous row has multiple outputs and current row just one input
            elif type_ == 'multi_2_single':
                for j, w in enumerate(children_widths[i - 1]):
                    subarrows.append(create_reverse_arrow(
                        0, sum(widths) - j + j * 4,
                        len(children_widths[i - 1]) - j + 1, j + 1
                    ))
                    widths.append(int(max_widths[j]))

            # if previous has single output and current row has single input
            elif type_ == 'single_2_single':
                subarrows.append(create_arrow(0, 0, 0, 0))

            # if previous row has one output and current row has multiple inputs
            else:
                assert type_ == 'single_2_multi'
                for j, w in enumerate(children_widths[i]):
                    if isinstance(t, Rollout):
                        subarrows.append(create_arrow(0, sum(widths) - j + j * 4,
                                                      len(children_widths[i]) - j, j + 1))
                    else:
                        subarrows.append(create_arrow(j, sum(widths) - j + j * 3,
                                                      len(children_widths[i]) - j, j + 1))
                    widths.append(int(max_widths[j]))

            # add signature names to the arrows
            tuple_to_str = lambda x: '(' + ', '.join([str(y) for y in x]) + ')'
            if (isinstance(t, Rollout) or isinstance(t, Parallel)) and not t._pd_name:
                all_params = []
                for tt in t.transforms:
                    all_params.append(list(tt._pd_get_signature().keys()))
                to_combine = [
                    ' ' * (sum(widths[:k + 1]) + 3 * k + 2) + tuple_to_str(params)
                    if len(params) > 1
                    else ' ' * (sum(widths[:k + 1]) + 3 * k + 2) + params[0]
                    for k, params in enumerate(all_params)
                ]
                to_format = combine_multi_line_strings(to_combine)
            else:
                params = [x for x in t._pd_get_signature()]
                to_format = '  ' + tuple_to_str(params) if len(params) > 1 else '  ' + params[0]
            to_format_pad_length = max([len(x.split('\n')) for x in subarrows]) - 1
            to_format = ''.join(['\n' for _ in range(to_format_pad_length)] + [to_format])

            # combine the arrows
            mark = combine_multi_line_strings(subarrows + [to_format])
            mark = '\n'.join(['   ' + x for x in mark.split('\n')])
            output.append(make_green(mark))
            output.append(make_bold(f'{i}: ') + r)
        return '\n'.join(output)

    def __call__(self, args):
        """Call method for Compose

        :param args: Arguments to call with.
        :return: Output from series of transforms.
        """
        for transform_ in self.transforms:
            args = transform_._pd_call_transform(args)
        return args

    def _pd_forward_part(self) -> Transform:
        t_list = []
        for transform_, component_set in zip(self.transforms, self._pd_component_list):
            if 'forward' in component_set:
                if len(component_set) == 1:
                    t_list.append(transform_)
                else:
                    t_list.append(transform_.pd_forward)

        if len(t_list) == 1:
            forward = t_list[0]
        elif t_list:
            forward = Compose(t_list, call_info=self._pd_call_info)
        else:
            forward = Identity()

        return forward

    def _pd_preprocess_part(self) -> Transform:
        t_list = []
        for transform_, component_set in zip(self.transforms, self._pd_component_list):
            if 'preprocess' in component_set:
                if len(component_set) == 1:
                    t_list.append(transform_)
                else:
                    t_list.append(transform_.pd_preprocess)

        if len(t_list) == 1:
            pre = t_list[0]
        elif t_list:
            pre = Compose(t_list, call_info=self._pd_call_info)
        else:
            pre = Identity()
        return pre

    def _pd_postprocess_part(self) -> Transform:
        t_list = []
        for transform_, component_set in zip(self.transforms, self._pd_component_list):
            if 'postprocess' in component_set:
                if len(component_set) == 1:
                    t_list.append(transform_)
                else:
                    t_list.append(transform_.pd_postprocess)

        if len(t_list) == 1:
            post = t_list[0]
        elif t_list:
            post = Compose(t_list, call_info=self._pd_call_info)
        else:
            post = Identity()
        return post


class Rollout(CompoundTransform):
    """Apply a list of transform to same input and get tuple output

    Rollout([t1, t2, ...])(x) := (t1(x), t2(x), ...)

    :param transforms: List of transforms to rollout.
    :param call_info: A `CallInfo` object containing information about the how the transform was
        created (needed for saving).
    :param pd_name: Name of the transform.
    :param pd_group: If *True*, do not flatten this when used as child transform in a
        `CompoundTransform`.
    """
    op = '+'
    display_op = '+'

    def __init__(self, transforms: Iterable[Transform], call_info: inspector.CallInfo = None,
                 pd_name: str = None, pd_group=False):
        super().__init__(transforms, call_info=call_info, pd_name=pd_name, pd_group=pd_group)
        self.pd_keys = self._pd_get_keys(self.transforms)
        self._pd_output_format = namedtuple('namedtuple', self.pd_keys)

    def __call__(self, args):
        """Call method for Rollout

        :param args: Argument to call with
        :return: namedtuple of outputs
        """
        out = []
        for transform_ in self.transforms:
            out.append(transform_._pd_call_transform(args))
        if Transform.pd_stage is not None:
            return tuple(out)
        return self._pd_output_format(*out)

    def _pd_preprocess_part(self) -> Transform:
        t_list = [x.pd_preprocess for x in self.transforms]
        if all([isinstance(t, Identity) for t in t_list]):
            pre = Identity()
        else:
            pre = Rollout(t_list, call_info=self._pd_call_info)
        return pre

    def _pd_forward_part(self) -> Transform:
        t_list = [x.pd_forward for x in self.transforms]
        if all([isinstance(t, Identity) for t in t_list]):
            forward = Identity()
        elif 'preprocess' in self._pd_component and 'forward' in self._pd_component:
            forward = Parallel(t_list, call_info=self._pd_call_info)
        else:
            forward = Rollout(t_list, call_info=self._pd_call_info)
        return forward

    def _pd_postprocess_part(self) -> Transform:
        t_list = [x.pd_postprocess for x in self.transforms]
        if all([isinstance(t, Identity) for t in t_list]):
            post = Identity()
        elif len(list(self._pd_component)) >= 2 and 'postprocess' in self._pd_component:
            post = Parallel(t_list, call_info=self._pd_call_info)
        else:
            post = Rollout(t_list, call_info=self._pd_call_info)
        return post

    def _pd_longrepr(self, formatting=True) -> str:
        make_green_ = lambda x: make_green(x, not formatting)
        make_bold_ = lambda x: make_bold(x, not formatting)
        between = f'\n{make_green_("│ " + self.display_op)}  \n'
        rows = [make_green_('├─▶ ') + make_bold_(f'{i}: ') + t._pd_shortrepr()
                for i, t in enumerate(self.transforms[:-1])]
        rows.append(make_green_('└─▶ ') + make_bold_(f'{len(self.transforms) - 1}: ')
                    + self.transforms[-1]._pd_shortrepr())
        return between.join(rows) + '\n'


class Parallel(CompoundTransform):
    """Apply transforms in parallel to a tuple of inputs and get tuple output

    Parallel([f1, f2, ...])((x1, x2, ..)) := (f1(x1), f2(x2), ...)

    :param transforms: List of transforms to parallelize.
    :param call_info: A `CallInfo` object containing information about the how the transform was
        created (needed for saving).
    :param pd_name: Name of the transform.
    :param pd_group: If *True*, do not flatten this when used as child transform in a
        `CompoundTransform`.
    """
    op = '/'
    display_op = '/'

    def __init__(self, transforms, call_info=None, pd_name=None, pd_group=False):
        super().__init__(transforms, call_info=call_info, pd_name=pd_name, pd_group=pd_group)
        self.pd_keys = self._pd_get_keys(self.transforms)
        self._pd_output_format = namedtuple('namedtuple', self.pd_keys)

    def __call__(self, args):
        """Call method for Parallel

        :param args: Argument to call with.
        :return: Namedtuple of output.
        """
        out = []
        for ind, transform_ in enumerate(self.transforms):
            out.append(transform_._pd_call_transform(args[ind]))
        if Transform.pd_stage is not None:
            return tuple(out)
        return self._pd_output_format(*out)

    def _pd_preprocess_part(self) -> Transform:
        t_list = [x.pd_preprocess for x in self.transforms]
        if all([isinstance(t, Identity) for t in t_list]):
            pre = Identity()
        else:
            pre = Parallel(t_list, call_info=self._pd_call_info)
        return pre

    def _pd_forward_part(self) -> Transform:
        t_list = [x.pd_forward for x in self.transforms]
        if all([isinstance(t, Identity) for t in t_list]):
            forward = Identity()
        else:
            forward = Parallel(t_list, call_info=self._pd_call_info)
        return forward

    def _pd_postprocess_part(self) -> Transform:
        t_list = [x.pd_postprocess for x in self.transforms]
        if all([isinstance(t, Identity) for t in t_list]):
            post = Identity()
        else:
            post = Parallel(t_list, call_info=self._pd_call_info)
        return post

    def _pd_longrepr(self, formatting=True) -> str:
        if not formatting:
            make_green_ = lambda x: x
            make_bold_ = lambda x: x
        else:
            make_green_ = make_green
            make_bold_ = make_bold
        def pipes(n):
            return "│" * n
        def spaces(n):
            return " " * n
        def horizontal(n):
            return "─" * n
        len_ = len(self.transforms)
        out = ''
        for i, t in enumerate(self.transforms):
            out += (
                make_green_(pipes(len_ - i - 1) + '└' + horizontal(i + 1) + '▶ ') +
                make_bold_(f'{i}: ') + t._pd_shortrepr() + '\n'
            )
            if i < len(self.transforms) - 1:
                out += f'{make_green_(pipes(len_ - i - 1) + spaces(i + 2) + self.display_op)}  \n'
        return out


class BuiltinTransform(AtomicTransform):
    def __init__(self, call):
        super().__init__(call)

    def _pd_build_codegraph(self, graph: Optional[dict] = None,
                            scopemap: Optional[dict] = None,
                            name: Optional[str] = None,
                            scope: Optional[symfinder.Scope] = None) -> Tuple[dict, dict]:
        if graph is None:
            graph = {}
        if scopemap is None:
            scopemap = {}

        if scope is None:
            scope = self._pd_call_info.scope

        if ('padl', scope) not in graph:
            emptyscope = symfinder.Scope.empty()
            graph['padl', emptyscope] = var2mod.CodeNode.from_source('import padl', scope)
            scopemap['padl', scope] = emptyscope

        if name is not None:
            start_source = f'{name or "_pd_dummy"} = {self._pd_evaluable_repr()}'
            graph[name, scope] = \
                var2mod.CodeNode.from_source(start_source, scope)

            scopemap[name, scope] = scope

        return graph, scopemap

    def _pd_longrepr(self, formatting=True):
        return self._pd_call.split('padl.')[-1]


class Identity(BuiltinTransform):
    """Do nothing. Just pass on."""

    def __init__(self):
        super().__init__('padl.Identity()')

    def __call__(self, args):
        return args


class Unbatchify(ClassTransform):
    """Mark start of postprocessing.

    Unbatchify removes batch dimension (inverse of Batchify) and moves the input tensors to 'cpu'.

    :param dim: Batching dimension.
    :param cpu: If *True*, moves output to cpu after unbatchify.
    """

    def __init__(self, dim=0, cpu=True):
        super().__init__(arguments=OrderedDict([('dim', dim), ('cpu', cpu)]))
        self.dim = dim
        self._pd_component = {'postprocess'}
        self.cpu = cpu

    def _move_to_device(self, args):
        if isinstance(args, (tuple, list)):
            return tuple([self._move_to_device(x) for x in args])
        if isinstance(args, torch.Tensor):
            return args.to('cpu')
        return args

    def __call__(self, args):
        assert Transform.pd_stage is not None, ('Stage is not set, use infer_apply, eval_apply '
                                                'or train_apply instead of calling the transform '
                                                'directly.')

        if Transform.pd_stage != 'infer':
            return self._move_to_device(args) if self.cpu else args
        if isinstance(args, tuple):
            return tuple([self(x) for x in args])
        if isinstance(args, torch.Tensor):
            args = args.squeeze(self.dim)
            return args.to('cpu') if self.cpu else args

        raise TypeError('only tensors and tuples of tensors recursively supported...')


class Batchify(ClassTransform):
    """Mark end of preprocessing.

    Batchify adds batch dimension at *dim*. During inference, this unsqueezes tensors and,
    recursively, tuples thereof. Batchify also moves the input tensors to device specified
    for the transform.

    :param dim: Batching dimension.
    """

    def __init__(self, dim=0):
        super().__init__(arguments=OrderedDict([('dim', dim)]))
        self.dim = dim
        self._pd_component = {'preprocess'}

    def __call__(self, args):
        assert Transform.pd_stage is not None, ('Stage is not set, use infer_apply, eval_apply '
                                                'or train_apply instead of calling the transform '
                                                'directly.')

        if Transform.pd_stage != 'infer':
            return args
        if isinstance(args, (tuple, list)):
            return tuple([self(x) for x in args])
        if isinstance(args, torch.Tensor):
            return args.unsqueeze(self.dim)
        if isinstance(args, (float, int)):
            return torch.tensor([args])
        raise TypeError('only tensors and tuples of tensors recursively supported...')


def save(transform: Transform, path: Union[Path, str], force_overwrite: bool = False):
    """Save the transform to a folder at *path*.

    The folder's name should end with '.padl'. If no extension is given, it will be added
    automatically.

    If the folder exist, call with *force_overwrite* = `True` to overwrite. Otherwise, this
    will raise a FileExistsError.
    """
    transform.pd_save(path, force_overwrite)


def load(path):
    """Load a transform (as saved with padl.save) from *path*. """
    path = Path(path)
    with open(path / 'transform.py') as f:
        source = f.read()
    module = types.ModuleType('lfload')
    module.__dict__.update({
        '_pd_source': source,
        '_pd_module': module,
        '__file__': str(path / 'transform.py')
    })
    code = compile(source, path/'transform.py', 'exec')
    exec(code, module.__dict__)
    transform = module._pd_main
    for i, subtrans in enumerate(transform._pd_all_transforms()):
        subtrans.pd_post_load(path, i)
    return transform


def group(transform: Union[Rollout, Parallel]):
    """Group transforms. This prevents them from being flattened when used

    Example:
    When writing a Rollout as `(a + (b + c))`, this is automatically flattened to `(a + b + c)`
    - i.e. the resulting Rollout transform expects a 3-tuple whose inputs are passed to `a`, `b`,
    `c` respectively. To prevent that, do (a + group(b + c)). The resulting Rollout will expect a
    2-tuple whose first item will be passed to `a` and whose second item will be passed to `b + c`.
    """
    return transform.grouped()



"""Data utilities. """


class SimpleDataset:
    """A simple dataset.

    Takes *samples* and applies *transform* to it.

    :param samples: An object implementing __getitem__ and __len__.
    :param transform: Preprocessing transform.
    :param exception: Exception to catch for (fall back to *default*).
    :param default: The default value to fall back to in case of exception.
    """

    def __init__(self, samples, transform, exception=None, default=None):
        self.samples = samples
        self.transform = transform
        self.exception = exception
        self.default = default

    def __getitem__(self, item):
        if self.exception:
            try:
                return self.transform(self.samples[item])
            except self.exception:
                return self.default
        return self.transform(self.samples[item])

    def __len__(self):
        return len(self.samples)



"""Utilities. """

from padl.dumptools import inspector
from padl.print_utils import format_argument
from padl.transforms import AtomicTransform


def _maketrans(attr, getitem=False):
    class T(AtomicTransform):
        """Dynamically generated transform for the "this" object.

        :param args: Arguments to pass to the input's method.
        :param kwargs: Keyword arguments to pass to the input's method.
        """

        def __init__(self, *args, **kwargs):
            self._args = args
            self._kwargs = kwargs
            caller_frameinfo = inspector.outer_caller_frameinfo(__name__)
            call_info = inspector.CallInfo(caller_frameinfo)
            if getitem:
                call = inspector.get_segment_from_frame(caller_frameinfo.frame, 'getitem')
            else:
                call = inspector.get_segment_from_frame(caller_frameinfo.frame, 'call')
            AtomicTransform.__init__(
                self,
                call=call,
                call_info=call_info,
            )

        def __call__(self, args):
            return getattr(args, attr)(*self._args, **self._kwargs)

        def _formatted_args(self) -> str:
            """Format the object's init arguments for printing. """
            args_list = [format_argument(val) for val in self._args]
            args_list += [f'{key}={format_argument(val)}' for key, val in self._kwargs.items()]
            return ', '.join(args_list)

        def _pd_longrepr(self, formatting=True):
            return self._pd_shortrepr()

        def _pd_shortrepr(self, formatting=True):
            return f'{attr}({self._formatted_args()})'

        def _pd_tinyrepr(self, formatting=True):
            return self.pd_name or attr

    return T


class _This:
    """Transform factory for capturing attributes/ get-items. """

    def __getitem__(self, item):
        return _maketrans('__getitem__', True)(item)

    def __getattr__(self, attr):
        return _maketrans(attr)


this = _This()



"""Utilities for finding packages used in code. """

import ast
import os
import sys
from importlib.metadata import version, PackageNotFoundError


def standard_lib_names_gen(include_underscored=False):
    standard_lib_dir = os.path.dirname(os.__file__)
    for filename in os.listdir(standard_lib_dir):
        if not include_underscored and filename.startswith('_'):
            continue
        filepath = os.path.join(standard_lib_dir, filename)
        name, ext = os.path.splitext(filename)
        if filename.endswith('.py') and os.path.isfile(filepath):
            if str.isidentifier(name):
                yield name
        elif os.path.isdir(filepath) and '__init__.py' in os.listdir(filepath):
            yield name


STDLIBNAMES = list(standard_lib_names_gen())


def get_packages(nodes):
    """Get a list of package names given a list of ast nodes *nodes*. """
    result = set()
    for node in nodes:
        if isinstance(node, ast.Import):
            for name in node.names:
                result.add(name.name.split('.')[0])
        if isinstance(node, ast.ImportFrom):
            result.add(node.module.split('.')[0])
    return result


def get_version(package):
    """Get a package's version (defaults to '?'). """
    try:
        return version(package)
    except PackageNotFoundError:
        return '?'


def dump_packages_versions(nodes):
    """Dump packages and their versions to a string.

    Format of the string is:

    <package>==<version>
    [...]

    :param nodes: List of ast nodes in a module.
    :returns: String with packages and versions.
    """
    result = f'# created with python-{".".join([str(x) for x in sys.version_info[:3]])}\n'
    for package in get_packages(nodes):
        if package in STDLIBNAMES:
            continue
        result += f'{package}=={get_version(package)}' + '\n'
    return result



from padl.dumptools.serialize import value
from padl.wrap import transform
from padl.transforms import Batchify, Unbatchify, Identity, group, load, save
from padl.util_transforms import IfTrain, IfEval, IfInfer, IfInStage
from padl.version import __version__
from padl.utils import this
identity = Identity()
batch = Batchify()
unbatch = Unbatchify()

__all__ = ['value', 'transform', 'Batchify', 'Unbatchify', 'Identity', 'group', 'load', 'save',
           'IfTrain', 'IfEval', 'IfInfer', 'IfInStage', 'identity', 'batch', 'unbatch', 'this',
           '__version__']



import pytest

from padl.dumptools import inspector


toplevel_callinfo = inspector.CallInfo('here')


def nested_builder():
    return inspector.CallInfo('here')


def double_nested_builder():
    def inner_builder():
        return inspector.CallInfo('here')
    return inner_builder()


class TestCallerInfo:
    def test_toplevel(self):
        assert toplevel_callinfo.module.__name__ == __name__
        assert toplevel_callinfo.scope.scopelist == []

    def test_nested(self):
        nested_callinfo = nested_builder()
        assert nested_callinfo.module.__name__ == __name__
        assert len(nested_callinfo.scope) == 1
        assert nested_callinfo.scope.scopelist[0][0] == 'nested_builder'

    def test_double_nested(self):
        with pytest.raises(AssertionError):
            double_nested_builder()


testcode = '''
a = 1
b = 2

def c():
    o = 7
    p = (a,
        b,
        c)

    f = a + \
        b

    k = f(1,
          2, 3,
          4)
    return f
'''


class TestGetStatement:
    def test_simple(self):
        statement, _ = inspector.get_statement(testcode, 2)
        assert statement == 'a = 1'
        statement, _ = inspector.get_statement(testcode, 3)
        assert statement == 'b = 2'

    def test_function(self):
        statement, _ = inspector.get_statement(testcode, 5)
        assert statement == testcode.split(' = 2')[1].strip()

    def test_multiline_1(self):
        statement, _ = inspector.get_statement(testcode, 7)
        assert statement == 'p = (a,\n    b,\n    c)'

    def test_multiline_2(self):
        statement, _ = inspector.get_statement(testcode, 8)
        assert statement == 'p = (a,\n    b,\n    c)'

    def test_multiline_3(self):
        statement, _ = inspector.get_statement(testcode, 9)
        assert statement == 'p = (a,\n    b,\n    c)'

    def test_multiline_4(self):
        statement, _ = inspector.get_statement(testcode, 11)
        assert statement == 'f = a +         b'

    def test_multiline_5(self):
        statement, _ = inspector.get_statement(testcode, 14)
        assert statement == 'k = f(1,\n      2, 3,\n      4)'



"""Exceptions."""


class WrongDeviceError(Exception):
    """Devices do not match for all transforms.

    :param mother_transform: Main transform that contains child transform.
    :param child_transform: Transform that is contained in mother transform.
    """

    def __init__(self, mother_transform, child_transform):
        super().__init__(f"{mother_transform._pd_shortrepr()} is in '{mother_transform.pd_device}',"
                         f' while {child_transform._pd_shortrepr()} is in '
                         f"'{child_transform.pd_device}'")



import ast
import json
from pathlib import Path
import sys

from padl.dumptools import inspector, sourceget, var2mod, symfinder


SCOPE = symfinder.Scope.toplevel(sys.modules[__name__])


class Serializer:
    """Serializer base class. """

    store = []
    i = 0

    def __init__(self):
        self.index = Serializer.i
        Serializer.i += 1
        self.store.append(self)

    @property
    def varname(self):
        """The varname to store in the dumped code. """
        return f'PADL_VALUE_{self.index}'

    @classmethod
    def save_all(cls, codegraph, scopemap, path):
        """Save all values. """
        for codenode in list(codegraph.values()):
            for serializer in cls.store:
                if serializer.varname in codenode.source:
                    codegraph.update(serializer.save(path))
                    for varname, scope in codenode.globals_:
                        if varname == serializer.varname:
                            scopemap[varname, scope] = SCOPE


class JSONSerializer(Serializer):
    """JSON Serializer (stores stuff as json).

    :param val: The value to serialize.
    """

    def __init__(self, val):
        self.val = val
        super().__init__()

    def save(self, path: Path):
        """Method for saving *self.val*. """
        if path is None:
            savepath = '?.json'
        else:
            savepath = path / f'{self.i}.json'
            with open(savepath, 'w') as f:
                json.dump(self.val, f)
        load_source = (
            f'with open(pathlib.Path(__file__).parent / \'{savepath.name}\') as f:\n'
            f'    {self.varname} = json.load(f)\n'
        )
        return {(self.varname, SCOPE): var2mod.CodeNode(source=load_source, globals_=set()),
                ('json', SCOPE): var2mod.CodeNode(source='import json', globals_=set(),
                                                  ast_node=ast.parse('import json').body[0]),
                ('pathlib', SCOPE): var2mod.CodeNode(source='import pathlib', globals_=set(),
                                                     ast_node=ast.parse('import pathlib').body[0])}


def _serialize(val):
    if hasattr(val, '__len__') and len(val) > 10:
        serializer = JSONSerializer(val)
        print('using json')
        return serializer.varname
    return repr(val)


def value(val):
    """Helper function that marks things in the code that should be stored by value. """
    caller_frameinfo = inspector.outer_caller_frameinfo(__name__)
    _call, locs = inspector.get_segment_from_frame(caller_frameinfo.frame, 'call', True)
    source = sourceget.get_source(caller_frameinfo.filename)
    sourceget.put_into_cache(caller_frameinfo.filename, sourceget.original(source),
                             _serialize(val), *locs)
    return val



"""Utilities for inspecting frames. """

import ast
import dis
import inspect
import sys
import types
from typing import Callable, Literal, Optional
from warnings import warn

from padl.dumptools import symfinder, var2mod
from padl.dumptools.sourceget import get_source, original, cut


class CallInfo:
    """Information about the calling context.

    Contains the following information:
        - the module from which the call was made
        - the function from which that call was made
        - the scope (see `symfinder.Scope`)

    :param origin: Where to look for the call, can be
        - "nextmodule": use the first frame not in the module the object was created in
        - "here": use the frame the object was created in
    :param drop_n: Drop *n* levels from the calling scope.
    :param ignore_scope: Don't try to determine the scope (use the toplevel scope instead).

    :returns: A CallInfo object.
    """

    def __init__(self, origin: Literal['nextmodule', 'here'] = 'nextmodule',
                 drop_n: int = 0, ignore_scope: bool = False):
        assert isinstance(origin, inspect.FrameInfo) or origin in ('nextmodule', 'here')
        if isinstance(origin, inspect.FrameInfo):
            caller_frameinfo = origin
        if origin == 'nextmodule':
            calling_module_name = inspect.currentframe().f_back.f_globals['__name__']
            caller_frameinfo = outer_caller_frameinfo(calling_module_name)
        elif origin == 'here':
            caller_frameinfo = inspect.stack()[1]
        self.function = caller_frameinfo.function
        self.scope = self._determine_scope(caller_frameinfo, drop_n, ignore_scope)

    def _determine_scope(self, caller_frameinfo: inspect.FrameInfo,
                         drop_n: int, ignore_scope: bool) -> symfinder.Scope:
        """Determine the scope of the caller frame. """
        module = _module(caller_frameinfo.frame)

        if self.function == '<module>' or ignore_scope:
            return symfinder.Scope.toplevel(module)
        try:
            call_source = get_segment_from_frame(caller_frameinfo.frame.f_back, 'call')
            definition_source = get_source(caller_frameinfo.filename)
            fdef_lineno = caller_frameinfo.frame.f_lineno
            calling_scope = symfinder.Scope.toplevel(_module(caller_frameinfo.frame.f_back))
            scope = symfinder.Scope.from_source(definition_source, fdef_lineno,
                                                call_source, module, drop_n,
                                                calling_scope)
            assert len(scope) <= 1, 'scope longer than 1 currently not supported'
            return scope
        except (SyntaxError, RuntimeError) as exc:
            warn(f'Error determining scope, using top level: {exc}')  # TODO: fix this
            return symfinder.Scope.toplevel(module)

    @property
    def module(self):
        """The calling module. """
        return self.scope.module


def non_init_caller_frameinfo() -> inspect.FrameInfo:
    """Get the FrameInfo for the first outer frame that is not of an "__init__" method. """
    stack = inspect.stack()
    frameinfo = None
    for frameinfo in stack[1:]:
        if frameinfo.function != '__init__':
            break
    assert frameinfo is not None and frameinfo.function != '__init__'
    return frameinfo


def trace_this(tracefunc: Callable, frame: Optional[types.FrameType] = None):
    """Call in a function body to trace the rest of the function execution with function
    *tracefunc*. *tracefunc* must match the requirements for the argument of `sys.settrace`
    (in the documentation of which more details can be found).

    Example:

    ```
    def tracefunc(frame, event, arg)
        if 'event' == 'return':
            print('returning', arg)

    def myfunction():
        [...]
        _trace_this(tracefunc)
        return 123
    ```

    :param tracefunc: Trace function (see documentation of `sys.settrace` for details).
    :param frame: The frame to trace (defaults to the caller's frame).
    """
    previous_tracefunc = sys.gettrace()

    if frame is None:
        # default is the caller's frame
        frame = inspect.currentframe().f_back

    def trace(frame, event, arg):
        tracefunc(frame, event, arg)
        if event == 'return':
            sys.settrace(previous_tracefunc)
        if previous_tracefunc is not None:
            previous_tracefunc(frame, event, arg)

    if previous_tracefunc is None:
        # set global tracefunc to something, this is required to enable local tracing
        sys.settrace(lambda _a, _b, _c: None)

    frame.f_trace = trace


def _instructions_up_to_call(x) -> list:
    """Get all instructions up to last CALL FUNCTION. """
    instructions = list(dis.get_instructions(x))
    i = 0
    for i, instruction in enumerate(instructions[::-1]):
        if instruction.opname.startswith('CALL_'):
            break
    return instructions[:-i]


def _instructions_in_name(x) -> list:
    """Get all instructions up to last CALL FUNCTION. """
    instructions = list(dis.get_instructions(x))
    i = 0
    for i, instruction in enumerate(instructions):
        if instruction.opname not in ('LOAD_NAME', 'LOAD_ATTR'):
            break
    return instructions[:i]


def _instructions_in_getitem(x) -> list:
    """Get all instructions up to last CALL FUNCTION. """
    instructions = list(dis.get_instructions(x))
    return instructions[:-1]


def _instructions_up_to_offset(x, lasti: int) -> list:
    """Get all instructions up to offset *lasti*. """
    instructions = []
    for instruction in dis.get_instructions(x):
        instructions.append(instruction)
        if instruction.offset == lasti:
            break
    return instructions


def get_statement(source: str, lineno: int):
    """Get complete (potentially multi-line) statement at line *lineno* out of *source*.

    :returns: A tuple of statement and offset. The offset is a tuple of row offset and col offset.
        It can be used to determine the location of the satement within the source.
    """
    for row_offset in range(lineno):
        try:
            block, lineno_in_block, col_offset = get_surrounding_block(source, lineno - row_offset)
        except ValueError:
            continue
        try:
            try:
                statement = _get_statement_from_block(block, lineno_in_block + row_offset)
                return statement, (lineno - row_offset - 1, -col_offset)
            except SyntaxError:
                statement = _get_statement_from_block('(\n' + block + '\n)',
                                                      lineno_in_block + row_offset + 1)
                return statement, (lineno - lineno_in_block - 1, -col_offset)
        except SyntaxError:
            continue
    raise SyntaxError("Couldn't find the statement.")


def _get_statement_from_block(block: str, lineno_in_block: int):
    """Get a statement from ."""
    module = ast.parse(block)
    stmts = []
    for stmt in module.body:
        if stmt.lineno <= lineno_in_block <= stmt.end_lineno:
            stmts.append(ast.get_source_segment(block, stmt))
    return '\n'.join(stmts)


def get_surrounding_block(source: str, lineno: int):
    """Get the code block surrounding the line at *lineno* in *source*.

    The code block surrounding a line is the largest block of lines with the same or larger
    indentation as the line itself.

    The result will be unindented.

    Raises a `ValueError` if the line at *lineno* is empty.

    :param source: The source to extract the block from.
    :param lineno: Number of the line for extracting the block.
    :returns: A tuple containing the block itself and the line number of the target line
        within the block and the number of spaces removed from the front.
    """
    lines = source.split('\n')
    before, after = lines[:lineno-1], lines[lineno:]
    white = _count_leading_whitespace(lines[lineno-1])
    if white is None:
        raise ValueError('Line is empty.')
    block = [lines[lineno-1][white:]]
    lineno_in_block = 1
    while before:
        next_ = before.pop(-1)
        next_white = _count_leading_whitespace(next_)
        if next_white is None or next_white >= white:
            block = [next_[white:]] + block
        else:
            break
        lineno_in_block += 1
    while after:
        next_ = after.pop(0)
        next_white = _count_leading_whitespace(next_)
        if next_white is None or next_white >= white:
            block = block + [next_[white:]]
        else:
            break
    return '\n'.join(block), lineno_in_block, white


def _module(frame: types.FrameType) -> types.ModuleType:
    """Get module of *frame*. """
    try:
        return frame.f_globals['_pd_module']
    except KeyError:
        return sys.modules[frame.f_globals['__name__']]


def outer_caller_frameinfo(module_name: str) -> inspect.FrameInfo:
    """Get the first level of the stack before entering the module with name *module_name*. """
    stack = inspect.stack()
    before = True
    for frameinfo in stack:
        if frameinfo.frame.f_globals['__name__'] == module_name:
            before = False
            continue
        if before:
            continue
        return frameinfo


def caller_module() -> types.ModuleType:
    """Get the first module of the caller. """
    calling_module_name = inspect.currentframe().f_back.f_globals['__name__']
    return _module(outer_caller_frameinfo(calling_module_name).frame)


def caller_frame() -> types.FrameType:
    """Get the callers frame. """
    calling_module_name = inspect.currentframe().f_back.f_globals['__name__']
    return outer_caller_frameinfo(calling_module_name).frame


def get_segment_from_frame(caller_frame: types.FrameType, segment_type, return_locs=False) -> str:
    """Get a segment of a given type from a frame.

    *NOTE*: All this is rather hacky and should be changed as soon as python 3.11 becomes widely
    available as then it will be possible to get column information from frames
    (see inline comments).

    *segement_type* can be 'call', 'attribute', 'getitem'.
    """
    if segment_type == 'call':
        node_type = ast.Call
        instructions_finder = _instructions_up_to_call
    elif segment_type == 'attribute':
        node_type = ast.Attribute
        instructions_finder = _instructions_in_name
    elif segment_type == 'getitem':
        node_type = ast.Subscript
        instructions_finder = _instructions_in_getitem
    # we want to extract the precise init statement here (e.g. `MyClass(1, 2, 3)`
    # , for python 3.11 (currently in development) this can be done via co_positions
    # (see https://www.python.org/dev/peps/pep-0657/),
    # for now, as 3.11 isn't widely used, this requires the following hack:
    # extract the source of the class init statement
    try:
        full_source = caller_frame.f_globals['_pd_source']
    except KeyError:
        full_source = get_source(caller_frame.f_code.co_filename)

    source, offset = get_statement(original(full_source), caller_frame.f_lineno)
    # the source can contain surrounding stuff we need to discard
    # as we only have the line number (this is what makes this complicated)

    # get all segments in the source that correspond to calls and might thus
    # potentially be the class init
    candidate_segments = var2mod.Finder(node_type).get_source_segments(source)
    # for each candidate, disassemble and compare the instructions to what we
    # actually have, a match means this is the correct statement
    if not candidate_segments:
        raise RuntimeError('No attributes found.')
    # disassemble and get the instructions up to the current position
    target_instrs = _instructions_up_to_offset(caller_frame.f_code,
                                               caller_frame.f_lasti)

    segment = None
    locs = None
    found = False

    for segment, locs in sorted(candidate_segments, key=lambda x: -len(x[0])):

        instrs = instructions_finder(segment)
        if len(instrs) > len(target_instrs):
            continue
        for instr, target_instr in zip(instrs, target_instrs[-len(instrs):]):
            if instr.argval != target_instr.argval:
                break
            same_opname = instr.opname == target_instr.opname
            load_ops = ('LOAD_NAME', 'LOAD_FAST', 'LOAD_GLOBAL')
            both_load = instr.opname in load_ops and target_instr.opname in load_ops
            if not (same_opname or both_load):
                break
        else:
            found = True
            break

    if segment is None or not found:
        raise RuntimeError(f'{segment_type} not found.')

    locs = (
        locs[0] - 1 + offset[0],
        locs[1] - 1 + offset[0],
        locs[2] - offset[1],
        locs[3] - offset[1]
    )
    # cutting is necessary instead of just using the segment from above for support of
    # `sourceget.ReplaceString`s
    segment = cut(full_source, *locs)

    if return_locs:
        return (
            segment, locs
        )
    return segment


def _count_leading_whitespace(line: str) -> int:
    """Count the number of spaces *line* starts with. """
    i = 0
    for char in line:
        if char == ' ':
            i += 1
            continue
        return i
